##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Tests of the distributed commit lock."""

import threading
import time

from persistent.TimeStamp import TimeStamp
from ZODB.Connection import TransactionMetaData
from ZODB.tests.StorageTestBase import MinPO
from ZODB.tests.StorageTestBase import zodb_pickle

from ZEO.Exceptions import ClientDisconnected
from ZEO.tests.TestThread import TestThread


ZERO = b'\0'*8


class WorkerThread(TestThread):

    # run the entire test in a thread so that the blocking call for
    # tpc_vote() doesn't hang the test suite.

    def __init__(self, test, storage, trans):
        self.storage = storage
        self.trans = trans
        self.ready = threading.Event()
        TestThread.__init__(self, test)

    def testrun(self):
        try:
            self.storage.tpc_begin(self.trans)
            oid = self.storage.new_oid()
            p = zodb_pickle(MinPO("c"))
            self.storage.store(oid, ZERO, p, '', self.trans)
            oid = self.storage.new_oid()
            p = zodb_pickle(MinPO("c"))
            self.storage.store(oid, ZERO, p, '', self.trans)
            self.myvote()
            self.storage.tpc_finish(self.trans)
        except ClientDisconnected:
            self.storage.tpc_abort(self.trans)
        except Exception:
            self.storage.tpc_abort(self.trans)
            raise

    def myvote(self):
        # The vote() call is synchronous, which makes it difficult to
        # coordinate the action of multiple threads that all call
        # vote().  This method sends the vote call, then sets the
        # event saying vote was called, then waits for the vote
        # response.

        future = self.storage._server.call('vote', id(self.trans), wait=False)
        self.ready.set()
        future.result(9)


class CommitLockTests:

    NUM_CLIENTS = 5

    # The commit lock tests verify that the storage successfully
    # blocks and restarts transactions when there is contention for a
    # single storage.  There are a lot of cases to cover.

    # The general flow of these tests is to start a transaction by
    # getting far enough into 2PC to acquire the commit lock.  Then
    # begin one or more other connections that also want to commit.
    # This causes the commit lock code to be exercised.  Once the
    # other connections are started, the first transaction completes.

    txn = None

    def _cleanup(self):
        if self.txn is not None:
            self._storage.tpc_abort(self.txn)
        for store, trans in self._storages:
            store.tpc_abort(trans)
            store.close()
        self._storages = []

    def _start_txn(self):
        txn = TransactionMetaData()
        self._storage.tpc_begin(txn)
        oid = self._storage.new_oid()
        self._storage.store(oid, ZERO, zodb_pickle(MinPO(1)), '', txn)
        self.txn = txn
        return oid, txn

    def _begin_threads(self):
        # Start a second transaction on a different connection without
        # blocking the test thread.  Returns only after each thread has
        # set it's ready event.
        self._storages = []
        self._threads = []
        self._exc = None

        for i in range(self.NUM_CLIENTS):
            storage = self._new_storage_client()
            txn = TransactionMetaData()
            self._get_timestamp()

            t = WorkerThread(self, storage, txn)
            self._threads.append(t)
            t.start()
            try:
                t.ready.wait(2)  # fail if this takes unreasonably long
            except Exception as exc:
                self._exc = exc
                return

            # Close one of the connections abnormally to test server response
            if i == 0:
                storage.close()
            else:
                self._storages.append((storage, txn))

    def _finish_threads(self):
        for t in self._threads:
            t.cleanup()

    def _get_timestamp(self):
        t = time.time()
        t = TimeStamp(*time.gmtime(t)[:5]+(t % 60,))
        return repr(t)


class CommitLockVoteTests(CommitLockTests):

    def checkCommitLockVoteFinish(self):
        oid, txn = self._start_txn()
        self._storage.tpc_vote(txn)

        self._begin_threads()
        self.assertIsNone(self._exc)

        self._storage.tpc_finish(txn)
        self._storage.load(oid, '')

        self._finish_threads()

        self._dostore()
        self._cleanup()

    def checkCommitLockVoteAbort(self):
        oid, txn = self._start_txn()
        self._storage.tpc_vote(txn)

        self._begin_threads()
        self.assertIsNone(self._exc)

        self._storage.tpc_abort(txn)

        self._finish_threads()

        self._dostore()
        self._cleanup()

    def checkCommitLockVoteClose(self):
        oid, txn = self._start_txn()
        self._storage.tpc_vote(txn)

        self._begin_threads()
        self.assertIsNone(self._exc)

        self._storage.close()

        self._finish_threads()
        self._cleanup()



"""ZEO server interface implementation."""

import asyncio
import json
import logging
import os
import threading

import ZODB.POSException

from ..shortrepr import short_repr
from . import base
from .compat import new_event_loop
from .marshal import encoder
from .marshal import reduce_exception
from .marshal import server_decoder


logger = logging.getLogger(__name__)


class ServerProtocol(base.ZEOBaseProtocol):
    """asyncio low-level ZEO server interface
    """

    protocols = (b'5', )

    methods = {'register'}

    unlogged_exception_types = (
        ZODB.POSException.POSKeyError,
        ZODB.POSException.ConflictError,
        ZODB.POSException.ReadConflictError,
        )

    def __init__(self, loop, addr, zeo_storage, msgpack):
        """Create a server's client interface
        """
        super().__init__(loop, repr(addr))
        self.addr = addr
        self.zeo_storage = zeo_storage

        self.announce_protocol = (
          (b'M' if msgpack else b'Z') + best_protocol_version
        )

    closed = False

    def close(self):
        logger.debug("Closing server protocol")
        if not self.closed:
            self.closed = True
            super().close()
            self.zeo_storage = None  # break reference cycle

    connected = None  # for tests

    def connection_made(self, transport):
        self.connected = True
        super().connection_made(transport)
        self.write_message(self.announce_protocol)

    def connection_lost(self, exc):
        super().connection_lost(exc)
        self.connected = False
        if exc:
            logger.error("Disconnected %s:%s", exc.__class__.__name__, exc)
        if not self.closed:  # unanticipated connection loss
            self.zeo_storage.notify_disconnected()
        self.stop()

    def stop(self):
        pass  # Might be replaced when running a thread per client

    def finish_connection(self, protocol_version):
        if protocol_version == b'ruok':
            self.write_message(
                json.dumps(self.zeo_storage.ruok()).encode("ascii"))
            self.close()
        else:
            version = protocol_version[1:]
            if version in self.protocols:
                logger.info("received handshake %r" %
                            str(protocol_version.decode('ascii')))
                self.protocol_version = protocol_version
                self.encode = encoder(protocol_version, True)
                self.decode = server_decoder(protocol_version)
                self.zeo_storage.notify_connected(self)
            else:
                logger.error("bad handshake %s" % short_repr(protocol_version))
                self.close()

    def call_soon_threadsafe(self, func, *args):
        try:
            self.loop.call_soon_threadsafe(func, *args)
        except RuntimeError:
            if self.connected:
                logger.exception("call_soon_threadsafe failed while connected")

    def message_received(self, message):
        try:
            message_id, async_, name, args = self.decode(message)
        except Exception:
            logger.exception("Can't deserialize message")
            self.close()
            return

        if message_id == -1:
            return  # keep-alive

        if name not in self.methods:
            logger.error('Invalid method, %r', name)
            self.close()

        try:
            result = getattr(self.zeo_storage, name)(*args)
        except Exception as exc:
            if not isinstance(exc, self.unlogged_exception_types):
                logger.exception(
                    "%s`%r` raised exception:",
                    'async ' if async_ else '', name)
            if async_:
                return self.close()  # No way to recover/cry for help
            else:
                return self.send_error(message_id, exc)

        if not async_:
            self.send_reply(message_id, result)

    def send_reply(self, message_id, result, send_error=False, flag=0):
        try:
            result = self.encode(message_id, flag, '.reply', result)
        except Exception:
            if isinstance(result, Delay):
                result.set_sender(message_id, self)
                return
            else:
                logger.exception("Unpicklable response %r", result)
                if not send_error:
                    self.send_error(
                        message_id,
                        ValueError("Couldn't pickle response"),
                        True)

        self.write_message(result)

    def send_reply_threadsafe(self, message_id, result):
        self.loop.call_soon_threadsafe(self.reply, message_id, result)

    def send_error(self, message_id, exc, send_error=False):
        """Abstracting here so we can make this cleaner in the future
        """
        self.send_reply(message_id, reduce_exception(exc), send_error, 2)

    def async_(self, method, *args):
        self.call_async(method, args)

    def async_threadsafe(self, method, *args):
        self.call_soon_threadsafe(self.call_async, method, args)


best_protocol_version = os.environ.get(
    'ZEO_SERVER_PROTOCOL',
    ServerProtocol.protocols[-1].decode('utf-8')).encode('utf-8')
assert best_protocol_version in ServerProtocol.protocols


def new_connection(loop, addr, socket, zeo_storage, msgpack):
    protocol = ServerProtocol(loop, addr, zeo_storage, msgpack)
    cr = loop.create_connection(lambda: protocol, sock=socket)
    asyncio.ensure_future(cr, loop=loop)


class Delay:
    """Used to delay response to client for synchronous calls.

    When a synchronous call is made and the original handler returns
    without handling the call, it returns a Delay object that prevents
    the mainloop from sending a response.
    """

    msgid = protocol = sent = None
    unlogged_exception_types = ServerProtocol.unlogged_exception_types

    def set_sender(self, msgid, protocol):
        self.msgid = msgid
        self.protocol = protocol

    def reply(self, obj):
        self.sent = 'reply'
        if self.protocol:
            self.protocol.send_reply(self.msgid, obj)

    def error(self, exc_info):
        self.sent = 'error'
        if exc_info[0] not in self.unlogged_exception_types:
            logger.error("Error raised in delayed method", exc_info=exc_info)
        if self.protocol:
            self.protocol.send_error(self.msgid, exc_info[1])

    def __repr__(self):
        return "{}[{}, {!r}, {!r}, {!r}]".format(
            self.__class__.__name__, id(self),
            self.msgid, self.protocol, self.sent)

    def __reduce__(self):
        raise TypeError("Can't pickle delays.")


class Result(Delay):

    def __init__(self, *args):
        self.args = args

    def set_sender(self, msgid, protocol):
        reply, callback = self.args
        protocol.send_reply(msgid, reply)
        callback()


class MTDelay(Delay):

    def __init__(self):
        self.ready = threading.Event()

    def set_sender(self, *args):
        Delay.set_sender(self, *args)
        self.ready.set()

    def reply(self, obj):
        self.ready.wait()
        self.protocol.call_soon_threadsafe(
            self.protocol.send_reply, self.msgid, obj)

    def error(self, exc_info):
        self.ready.wait()
        self.protocol.call_soon_threadsafe(Delay.error, self, exc_info)


class Acceptor:

    def __init__(self, storage_server, addr, ssl, msgpack):
        self.storage_server = storage_server
        self.addr = addr
        self.ssl_context = ssl
        self.msgpack = msgpack
        self.event_loop = loop = new_event_loop()

        if isinstance(addr, tuple):
            cr = loop.create_server(self.factory, addr[0], addr[1],
                                    reuse_address=True, ssl=ssl)
        else:
            cr = loop.create_unix_server(self.factory, addr, ssl=ssl)

        f = asyncio.ensure_future(cr, loop=loop)
        server = loop.run_until_complete(f)

        self.server = server
        if isinstance(addr, tuple) and addr[1] == 0:
            addrs = [s.getsockname() for s in server.sockets]
            addrs = [a for a in addrs if len(a) == len(addr)]
            if addrs:
                self.addr = addrs[0]
            else:
                self.addr = server.sockets[0].getsockname()[:len(addr)]

        logger.info("listening on %s", str(addr))

    def factory(self):
        try:
            logger.debug("Accepted connection")
            zs = self.storage_server.create_client_handler()
            protocol = ServerProtocol(
                self.event_loop, self.addr, zs, self.msgpack)
        except Exception:
            logger.exception("Failure in protocol factory")

        return protocol

    def loop(self, timeout=None):
        self.event_loop.run_forever()
        self.event_loop.close()

    closed = False

    def close(self):
        if not self.closed:
            self.closed = True
            self.event_loop.call_soon_threadsafe(self._close)

    def _close(self):
        loop = self.event_loop

        self.server.close()

        f = asyncio.ensure_future(self.server.wait_closed(), loop=loop)

        @f.add_done_callback
        def server_closed(f):
            # stop the loop when the server closes:
            loop.call_soon(loop.stop)

        def timeout():
            logger.warning("Timed out closing asyncio.Server")
            loop.call_soon(loop.stop)

        # But if the server doesn't close in a second, stop the loop anyway.
        loop.call_later(1, timeout)

        self.server = None  # break reference cycle



# cython: language_level=3, boundscheck=False, wraparound=False

"""``cython`` implementation for ``SizedMessageProtocol``."""
from cpython.bytes cimport PyBytes_FromStringAndSize
from libc.string cimport memcpy

import logging
import struct


logger = logging.getLogger(__name__)


cdef enum ReadState:
    process_size = 1
    process_message = 2
    closed = 0

cdef object pack = struct.pack
cdef object unpack = struct.unpack


cdef class SizedMessageProtocol:
    cdef object __closed
    cdef public object receive  # callback for received messages
    cdef public object connection_lost_called

    def __init__(self, receive):
        self.receive = receive
        self.__closed = self.connection_lost_called = False

    def set_receive(self, receive):
        self.receive = receive

    cdef public object transport

    def close(self):
        if self.__closed:
            return
        self.__closed = True
        self.eof_received()
        self.transport.close()
        # break reference cycles
        self.transport = self.receive = self.writelines = None

    # output
    cdef bint paused
    cdef list output  # buffer
    cdef object writelines

    cdef _write_message(self, bytes message):
        self.writelines((pack(">I", len(message)), message))

    def write_message(self, message):
        if self.paused:
            self.output.append(message)
        else:
            self._write_message(message)

    def write_message_iter(self, message_iter):
        it = iter(message_iter)
        if self.paused:
            self.output.append(it)
            return
        for message in it:
            self._write_message(message)
            if self.paused:
                self.output.append(it)
                return

    # protocol responsibilities
    def pause_writing(self):
        self.paused = 1

    def resume_writing(self):
        self.paused = 0
        cdef list output = self.output
        while output and not self.paused:
            message = output.pop(0)
            if type(message) is bytes:
                self._write_message(message)
            else:
                it = message
                for message in it:
                    self._write_message(message)
                    if self.paused:
                        self.output.insert(0, it)
                        return

    # input
    cdef ReadState read_state  # current read state
    cdef unsigned read_wanted  # wanted data size
    cdef unsigned received_count  # received unprocessed bytes
    cdef list chunk_buffer  # received data chunks
    cdef unsigned chunk_index  # unprocessed index in 1. chunk

    # protocol responsibilities
    def data_received(self, bytes data):
        self.chunk_buffer.append(data)
        self.received_count += len(data)
        cdef unsigned wanted = self.read_wanted
        cdef bytes target
        cdef unsigned char *tv
        cdef unsigned tvi
        cdef bytes chunk
        cdef const unsigned char *cv
        cdef unsigned ci
        cdef unsigned unprocessed, use, i
        while self.read_state and self.read_wanted <= self.received_count:
            wanted = self.read_wanted
            tv = target = PyBytes_FromStringAndSize(<char *> NULL, wanted)
            tvi = 0
            while wanted:
                cv = chunk = self.chunk_buffer[0]
                ci = self.chunk_index
                unprocessed = len(chunk) - ci
                if unprocessed > wanted:
                    use = wanted
                    self.chunk_index += wanted
                else:
                    use = unprocessed
                    self.chunk_buffer.pop(0)
                    self.chunk_index = 0
                if use <= 4:
                    for i in range(use):
                        tv[tvi + i] = cv[ci + i]
                else:
                    memcpy(&tv[tvi], &cv[ci], use)
                tvi += use
                wanted -= use
            self.received_count -= self.read_wanted
            if self.read_state == process_size:
                self.read_state = process_message
                self.read_wanted = unpack(">I", target)[0]
            else:  # read_state == process_message
                try:
                    self.receive(target)
                except Exception:
                    logger.exception("Processing message `%r` failed"
                                     % target)
                if self.read_state:  # not yet closed
                    self.read_state = process_size
                    self.read_wanted = 4

    def connection_made(self, transport):
        self.transport = transport
        self.writelines = transport.writelines
        self.paused = 0
        self.output = []
        self.read_state = process_size
        self.read_wanted = 4
        self.received_count = 0
        self.chunk_buffer = []
        self.chunk_index = 0


    def connection_lost(self, exc):
        self.connection_lost_called = True
        if self.__closed:
            return
        self.transport.close()

    def eof_received(self):
        self.read_state = closed



#!/usr/bin/env python2.3

"""Make sure a ZEO server is running.

usage: zeoup.py [options]

The test will connect to a ZEO server, load the root object, and attempt to
update the zeoup counter in the root.  It will report success if it updates
the counter or if it gets a ConflictError.  A ConflictError is considered a
success, because the client was able to start a transaction.

Options:

    -p port -- port to connect to

    -h host -- host to connect to (default is current host)

    -S storage -- storage name (default '1')

    -U path -- Unix-domain socket to connect to

    --nowrite -- Do not update the zeoup counter.

    -1 -- Connect to a ZEO 1.0 server.

You must specify either -p and -h or -U.
"""

import getopt
import logging
import socket
import sys
import time

import transaction
import ZODB
from persistent.mapping import PersistentMapping
from ZODB.POSException import ConflictError
from ZODB.tests.MinPO import MinPO

from ZEO.ClientStorage import ClientStorage


ZEO_VERSION = 2


def setup_logging():
    # Set up logging to stderr which will show messages originating
    # at severity ERROR or higher.
    root = logging.getLogger()
    root.setLevel(logging.ERROR)
    fmt = logging.Formatter(
        "------\n%(asctime)s %(levelname)s %(name)s %(message)s",
        "%Y-%m-%dT%H:%M:%S")
    handler = logging.StreamHandler()
    handler.setFormatter(fmt)
    root.addHandler(handler)


def check_server(addr, storage, write):
    t0 = time.time()
    if ZEO_VERSION == 2:
        # TODO:  should do retries w/ exponential backoff.
        cs = ClientStorage(addr, storage=storage, wait=0,
                           read_only=(not write))
    else:
        cs = ClientStorage(addr, storage=storage, debug=1,
                           wait_for_server_on_startup=1)
    # _startup() is an artifact of the way ZEO 1.0 works.  The
    # ClientStorage doesn't get fully initialized until registerDB()
    # is called.  The only thing we care about, though, is that
    # registerDB() calls _startup().

    if write:
        db = ZODB.DB(cs)
        cn = db.open()
        root = cn.root()
        try:
            # We store the data in a special `monitor' dict under the root,
            # where other tools may also store such heartbeat and bookkeeping
            # type data.
            monitor = root.get('monitor')
            if monitor is None:
                monitor = root['monitor'] = PersistentMapping()
            obj = monitor['zeoup'] = monitor.get('zeoup', MinPO(0))
            obj.value += 1
            transaction.commit()
        except ConflictError:
            pass
        cn.close()
        db.close()
    else:
        data, serial = cs.load("\0\0\0\0\0\0\0\0", "")
        cs.close()
    t1 = time.time()
    print("Elapsed time: %.2f" % (t1 - t0))


def usage(exit=1):
    print(__doc__)
    print(" ".join(sys.argv))
    sys.exit(exit)


def main():
    host = None
    port = None
    unix = None
    write = 1
    storage = '1'
    try:
        opts, args = getopt.getopt(sys.argv[1:], 'p:h:U:S:1',
                                   ['nowrite'])
        for o, a in opts:
            if o == '-p':
                port = int(a)
            elif o == '-h':
                host = a
            elif o == '-U':
                unix = a
            elif o == '-S':
                storage = a
            elif o == '--nowrite':
                write = 0
            elif o == '-1':
                ZEO_VERSION = 1  # NOQA: F841 unused variable
    except Exception as err:
        s = str(err)
        if s:
            s = ": " + s
        print(err.__class__.__name__ + s)
        usage()

    if unix is not None:
        addr = unix
    else:
        if host is None:
            host = socket.gethostname()
        if port is None:
            usage()
        addr = host, port

    setup_logging()
    check_server(addr, storage, write)


if __name__ == "__main__":
    try:
        main()
    except SystemExit:
        raise
    except Exception as err:
        s = str(err)
        if s:
            s = ": " + s
        print(err.__class__.__name__ + s)
        sys.exit(1)



##############################################################################
#
# Copyright (c) 2004 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import doctest
import re
import unittest

from zope.testing import renormalizing


def test_suite():
    return unittest.TestSuite((
        doctest.DocFileSuite(
            'zeopack.test',
            checker=renormalizing.RENormalizing([
                (re.compile('usage: Usage: '), 'Usage: '),  # Py 2.4
                (re.compile('options:'), 'Options:'),  # Py 2.4
                ]),
            ),
        ))



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Test suite for ZEO based on ZODB.tests."""
import doctest
import logging
import multiprocessing
import os
import pprint
import re
import resource
import shutil
import signal
import stat
import sys
import tempfile
import threading
import time
import unittest

import persistent
import transaction
import ZODB
import ZODB.blob
import ZODB.tests.hexstorage
import ZODB.tests.testblob
import ZODB.tests.util
import ZODB.utils
import zope.testing.setupstack
from ZODB.Connection import TransactionMetaData
from ZODB.tests import BasicStorage
from ZODB.tests import ConflictResolution
from ZODB.tests import IteratorStorage
from ZODB.tests import MTStorage
from ZODB.tests import PackableStorage
from ZODB.tests import ReadOnlyStorage
from ZODB.tests import RecoveryStorage
from ZODB.tests import RevisionStorage
from ZODB.tests import StorageTestBase
from ZODB.tests import Synchronization
from ZODB.tests import TransactionalUndoStorage
from ZODB.tests.MinPO import MinPO
from ZODB.tests.StorageTestBase import zodb_unpickle
from ZODB.utils import maxtid
from ZODB.utils import p64
from ZODB.utils import u64
from ZODB.utils import z64
from zope.testing import renormalizing

import ZEO.StorageServer
import ZEO.tests.ConnectionTests
from ZEO._compat import WIN
from ZEO.ClientStorage import ClientStorage
from ZEO.Exceptions import ClientDisconnected
from ZEO.tests import Cache
from ZEO.tests import CommitLockTests
from ZEO.tests import IterationTests
from ZEO.tests import ThreadTests
from ZEO.tests import forker

from . import testssl


logger = logging.getLogger('ZEO.tests.testZEO')


class DummyDB:

    def invalidate(self, *args):
        pass

    def invalidateCache(*unused):
        pass

    transform_record_data = untransform_record_data = lambda self, v: v


class CreativeGetState(persistent.Persistent):
    def __getstate__(self):
        self.name = 'me'
        return super().__getstate__()


class Test_convenience_functions(unittest.TestCase):

    def test_ZEO_client_convenience(self):
        from unittest import mock

        import ZEO

        client_thread = mock.Mock(
            spec=['call', 'async', 'async_iter', 'wait', 'close'])
        client = ZEO.client(
            8001, wait=False, _client_factory=client_thread)
        self.assertIsInstance(client, ClientStorage)
        client.close()
        client._cache.close()  # client thread responsibility

    def test_ZEO_DB_convenience_ok(self):
        from unittest import mock

        import ZEO

        client_mock = mock.Mock(spec=['close'])
        client_patch = mock.patch('ZEO.client', return_value=client_mock)
        DB_patch = mock.patch('ZODB.DB')

        dummy = object()

        with client_patch as client:
            with DB_patch as patched:
                db = ZEO.DB(dummy)

        self.assertIs(db, patched())
        client.assert_called_once_with(dummy)
        client_mock.close.assert_not_called()

    def test_ZEO_DB_convenience_error(self):
        from unittest import mock

        import ZEO

        client_mock = mock.Mock(spec=['close'])
        client_patch = mock.patch('ZEO.client', return_value=client_mock)
        DB_patch = mock.patch('ZODB.DB', side_effect=ValueError)

        dummy = object()

        with client_patch as client:
            with DB_patch:
                with self.assertRaises(ValueError):
                    ZEO.DB(dummy)

        client.assert_called_once_with(dummy)
        client_mock.close.assert_called_once()

    def test_ZEO_connection_convenience_ok(self):
        from unittest import mock

        import ZEO

        ret = object()
        DB_mock = mock.Mock(spec=[
            'close', 'open_then_close_db_when_connection_closes'])
        DB_mock.open_then_close_db_when_connection_closes.return_value = ret
        DB_patch = mock.patch('ZEO.DB', return_value=DB_mock)

        dummy = object()

        with DB_patch as patched:
            conn = ZEO.connection(dummy)

        self.assertIs(conn, ret)
        patched.assert_called_once_with(dummy)
        DB_mock.close.assert_not_called()

    def test_ZEO_connection_convenience_value(self):
        from unittest import mock

        import ZEO

        DB_mock = mock.Mock(spec=[
            'close', 'open_then_close_db_when_connection_closes'])
        otc = DB_mock.open_then_close_db_when_connection_closes
        otc.side_effect = ValueError
        DB_patch = mock.patch('ZEO.DB', return_value=DB_mock)

        dummy = object()

        with DB_patch as patched:
            with self.assertRaises(ValueError):
                ZEO.connection(dummy)

        patched.assert_called_once_with(dummy)
        DB_mock.close.assert_called_once()


class MiscZEOTests:
    """ZEO tests that don't fit in elsewhere."""

    def checkCreativeGetState(self):
        # This test covers persistent objects that provide their own
        # __getstate__ which modifies the state of the object.
        # For details see bug #98275

        db = ZODB.DB(self._storage)
        cn = db.open()
        rt = cn.root()
        m = CreativeGetState()
        m.attr = 'hi'
        rt['a'] = m

        # This commit used to fail because of the `Mine` object being put back
        # into `changed` state although it was already stored causing the ZEO
        # cache to bail out.
        transaction.commit()
        cn.close()

    def checkLargeUpdate(self):
        obj = MinPO("X" * (10 * 128 * 1024))
        self._dostore(data=obj)

    def checkZEOInvalidation(self):
        storage2 = self._new_storage_client()
        try:
            oid = self._storage.new_oid()
            ob = MinPO('first')
            revid1 = self._dostore(oid, data=ob)
            data, serial = storage2.load(oid, '')
            self.assertEqual(zodb_unpickle(data), MinPO('first'))
            self.assertEqual(serial, revid1)
            revid2 = self._dostore(oid, data=MinPO('second'), revid=revid1)

            # Now, storage 2 should eventually get the new data. It
            # will take some time, although hopefully not much.
            # We'll poll till we get it and whine if we time out:
            for n in range(30):
                time.sleep(.1)
                data, serial = storage2.load(oid, '')
                if serial == revid2 and \
                   zodb_unpickle(data) == MinPO('second'):
                    break
            else:
                raise AssertionError('Invalidation message was not sent!')
        finally:
            storage2.close()

    def checkVolatileCacheWithImmediateLastTransaction(self):
        # Earlier, a ClientStorage would not have the last transaction id
        # available right after successful connection, this is required now.
        storage2 = self._new_storage_client()
        self.assertTrue(storage2.is_connected())
        self.assertEqual(ZODB.utils.z64, storage2.lastTransaction())
        storage2.close()

        self._dostore()
        storage3 = self._new_storage_client()
        self.assertTrue(storage3.is_connected())
        self.assertEqual(8, len(storage3.lastTransaction()))
        self.assertNotEqual(ZODB.utils.z64, storage3.lastTransaction())
        storage3.close()


class GenericTestBase(
        # Base class for all ZODB tests
        StorageTestBase.StorageTestBase):

    shared_blob_dir = False
    blob_cache_dir = None
    server_debug = False

    def setUp(self):
        # Some operating systems like macOS set a very low soft limit for
        # the maximum number of open files. Some ZEO tests hit these limits.
        (soft_max_open_files,
         hard_max_open_files) = resource.getrlimit(resource.RLIMIT_NOFILE)
        if soft_max_open_files < 512:
            resource.setrlimit(resource.RLIMIT_NOFILE,
                               (512, hard_max_open_files))

        StorageTestBase.StorageTestBase.setUp(self)
        logger.info("setUp() %s", self.id())
        zport, stop = forker.start_zeo_server(
            self.getConfig(), self.getZEOConfig(), debug=self.server_debug)
        self._servers = [stop]
        if not self.blob_cache_dir:
            # This is the blob cache for ClientStorage
            self.blob_cache_dir = tempfile.mkdtemp(
                'blob_cache',
                dir=os.path.abspath(os.getcwd()))
        self._storage = self._wrap_client(
            ClientStorage(
                zport, '1', cache_size=20000000,
                min_disconnect_poll=0.5, wait=1,
                wait_timeout=60, blob_dir=self.blob_cache_dir,
                shared_blob_dir=self.shared_blob_dir,
                **self._client_options()),
            )
        self._storage.registerDB(DummyDB())

    # _new_storage_client opens another ClientStorage to the same storage
    # server self._storage is connected to. It is used by both ZEO and ZODB
    # tests.
    def _new_storage_client(self):
        client = ZEO.ClientStorage.ClientStorage(
            self._storage._addr, wait=1, **self._client_options())
        client = self._wrap_client(client)
        client.registerDB(DummyDB())
        return client

    def getZEOConfig(self):
        return forker.ZEOConfig(('127.0.0.1', 0))

    def _wrap_client(self, client):
        return client

    def _client_options(self):
        return {}

    def tearDown(self):
        self._storage.close()
        for stop in self._servers:
            stop()
        StorageTestBase.StorageTestBase.tearDown(self)


class GenericTests(
        GenericTestBase,
        # ZODB test mixin classes (in the same order as imported)
        BasicStorage.BasicStorage,
        PackableStorage.PackableStorage,
        Synchronization.SynchronizedStorage,
        MTStorage.MTStorage,
        ReadOnlyStorage.ReadOnlyStorage,
        # ZEO test mixin classes (in the same order as imported)
        CommitLockTests.CommitLockVoteTests,
        ThreadTests.ThreadTests,
        # Locally defined (see above)
        MiscZEOTests):
    """Combine tests from various origins in one class.
    """

    def open(self, read_only=0):
        # Needed to support ReadOnlyStorage tests.  Ought to be a
        # cleaner way.
        addr = self._storage._addr
        self._storage.close()
        self._storage = ClientStorage(
            addr, read_only=read_only, wait=1, **self._client_options())

    def checkWriteMethods(self):
        # ReadOnlyStorage defines checkWriteMethods.  The decision
        # about where to raise the read-only error was changed after
        # Zope 2.5 was released.  So this test needs to detect Zope
        # of the 2.5 vintage and skip the test.

        # The __version__ attribute was not present in Zope 2.5.
        if hasattr(ZODB, "__version__"):
            ReadOnlyStorage.ReadOnlyStorage.checkWriteMethods(self)

    def checkSortKey(self):
        key = f'{self._storage._storage}:{self._storage._server_addr}'
        self.assertEqual(self._storage.sortKey(), key)

    def _do_store_in_separate_thread(self, oid, revid, voted):

        def do_store():
            self.exception = None
            store = self._new_storage_client()
            try:
                t = transaction.get()
                self.assertEqual(store._connection_generation, 1)
                store.tpc_begin(t)
                self.assertEqual(store._connection_generation, 1)
                store.store(oid, revid, b'x', '', t)
                store.tpc_vote(t)
                store.tpc_finish(t)
            except Exception as v:
                self.exception = v
            finally:
                store.close()

        thread = threading.Thread(name='T2', target=do_store)
        thread.daemon = True
        thread.start()
        thread.join(voted and .1 or 9)
        if self.exception is not None:
            raise self.exception.with_traceback(self.exception.__traceback__)
        return thread


class FullGenericTests(
        GenericTests,
        Cache.TransUndoStorageWithCache,
        ConflictResolution.ConflictResolvingStorage,
        ConflictResolution.ConflictResolvingTransUndoStorage,
        PackableStorage.PackableUndoStorage,
        RevisionStorage.RevisionStorage,
        TransactionalUndoStorage.TransactionalUndoStorage,
        IteratorStorage.IteratorStorage,
        IterationTests.IterationTests):
    """Extend GenericTests with tests that MappingStorage can't pass."""

    def testPackUndoLog(self):
        # Prevent execution of the test inherited from ``ZODB>=6.0``.
        #
        # An adapted version of this test is executed via the following
        # ``checkPackUndoLog``.
        #
        # Once support for ``ZODB<6.0`` is dropped, this function
        # can go away and ``checkPackUndoLog`` can get renamed
        # to ``testPackUndoLog``.
        pass

    def checkPackUndoLog(self):
        # PackableStorage.PackableUndoStorage wants to adjust
        # time.sleep and time.time to cooperate and pretend for time
        # to pass. That doesn't work for the spawned server, and this
        # test case is very sensitive to times matching.
        super_meth = getattr(super(), "testPackUndoLog", None)
        if super_meth is None:  # old ZODB
            super_meth = super().checkPackUndoLog
        # Find the underlying function, not the decorated method.
        # If it doesn't exist, the implementation has changed and we
        # need to revisit this...
        underlying_func = super_meth.__wrapped__
        underlying_func(self)


class FileStorageRecoveryTests(StorageTestBase.StorageTestBase,
                               RecoveryStorage.RecoveryStorage):

    def getConfig(self):
        return """\
        <filestorage 1>
        path %s
        </filestorage>
        """ % tempfile.mktemp(dir='.')

    def _new_storage(self):
        zconf = forker.ZEOConfig(('127.0.0.1', 0))
        zport, stop = forker.start_zeo_server(self.getConfig(),
                                              zconf)
        self._servers.append(stop)

        blob_cache_dir = tempfile.mkdtemp(dir='.')

        storage = ClientStorage(
            zport, '1', cache_size=20000000,
            min_disconnect_poll=0.5, wait=1,
            wait_timeout=60, blob_dir=blob_cache_dir)
        storage.registerDB(DummyDB())
        return storage

    def setUp(self):
        StorageTestBase.StorageTestBase.setUp(self)
        self._servers = []

        self._storage = self._new_storage()
        self._dst = self._new_storage()

    def tearDown(self):
        self._storage.close()
        self._dst.close()

        for stop in self._servers:
            stop()
        StorageTestBase.StorageTestBase.tearDown(self)

    def new_dest(self):
        return self._new_storage()


class FileStorageTests(FullGenericTests):
    """Test ZEO backed by a FileStorage."""

    def getConfig(self):
        return """\
        <filestorage 1>
        path Data.fs
        </filestorage>
        """

    _expected_interfaces = (
        ('ZODB.interfaces', 'IStorageRestoreable'),
        ('ZODB.interfaces', 'IStorageIteration'),
        ('ZODB.interfaces', 'IStorageUndoable'),
        ('ZODB.interfaces', 'IStorageCurrentRecordIteration'),
        ('ZODB.interfaces', 'IExternalGC'),
        ('ZODB.interfaces', 'IStorage'),
        ('zope.interface', 'Interface'),
        )

    def checkInterfaceFromRemoteStorage(self):
        # ClientStorage itself doesn't implement IStorageIteration, but the
        # FileStorage on the other end does, and thus the ClientStorage
        # instance that is connected to it reflects this.
        self.assertFalse(ZODB.interfaces.IStorageIteration.implementedBy(
            ZEO.ClientStorage.ClientStorage))
        self.assertTrue(ZODB.interfaces.IStorageIteration.providedBy(
            self._storage))
        # This is communicated using ClientStorage's _info object:
        self.assertEqual(self._expected_interfaces,
                         self._storage._info['interfaces'])


class FileStorageSSLTests(FileStorageTests):

    def getZEOConfig(self):
        return testssl.server_config

    def _client_options(self):
        return {'ssl': testssl.client_ssl()}


class FileStorageHexTests(FileStorageTests):
    _expected_interfaces = (
        ('ZODB.interfaces', 'IStorageRestoreable'),
        ('ZODB.interfaces', 'IStorageIteration'),
        ('ZODB.interfaces', 'IStorageUndoable'),
        ('ZODB.interfaces', 'IStorageCurrentRecordIteration'),
        ('ZODB.interfaces', 'IExternalGC'),
        ('ZODB.interfaces', 'IStorage'),
        ('ZODB.interfaces', 'IStorageWrapper'),
        ('zope.interface', 'Interface'),
        )

    def getConfig(self):
        return """\
        %import ZODB.tests
        <hexstorage>
        <filestorage 1>
        path Data.fs
        </filestorage>
        </hexstorage>
        """


class FileStorageClientHexTests(FileStorageHexTests):

    use_extension_bytes = True

    def getConfig(self):
        return """\
        %import ZODB.tests
        <serverhexstorage>
        <filestorage 1>
        path Data.fs
        </filestorage>
        </serverhexstorage>
        """

    def _wrap_client(self, client):
        return ZODB.tests.hexstorage.HexStorage(client)


class FileStorageLoadDelayedTests(FileStorageTests):
    """Test ZEO backed by a FileStorage with delayes injected after load
       operations.

       This catches e.g. races in between loads and invalidations.
       See https://github.com/zopefoundation/ZEO/issues/209 for details.
    """

    level = 10  # very long test

    def getConfig(self):
        return """\
        %import ZEO.tests
        <loaddelayed_storage>
        <filestorage 1>
        path Data.fs
        </filestorage>
        </loaddelayed_storage>
        """


class ClientConflictResolutionTests(
        GenericTestBase,
        ConflictResolution.ConflictResolvingStorage):

    def getConfig(self):
        return '<mappingstorage>\n</mappingstorage>\n'

    def getZEOConfig(self):
        # Using '' can result in binding to :: and cause problems
        # connecting to the MTAcceptor on Travis CI
        return forker.ZEOConfig(('127.0.0.1', 0),
                                client_conflict_resolution=True)


class MappingStorageTests(GenericTests):
    """ZEO backed by a Mapping storage."""

    def getConfig(self):
        return """<mappingstorage 1/>"""

    def checkSimpleIteration(self):
        # The test base class IteratorStorage assumes that we keep undo data
        # to construct our iterator, which we don't, so we disable this test.
        pass

    def checkUndoZombie(self):
        # The test base class IteratorStorage assumes that we keep undo data
        # to construct our iterator, which we don't, so we disable this test.
        pass


class DemoStorageTests(GenericTests):

    def getConfig(self):
        return """
        <demostorage 1>
          <filestorage 1>
             path Data.fs
          </filestorage>
        </demostorage>
        """

    def checkUndoZombie(self):
        # The test base class IteratorStorage assumes that we keep undo data
        # to construct our iterator, which we don't, so we disable this test.
        pass

    def checkPackWithMultiDatabaseReferences(self):
        pass  # DemoStorage pack doesn't do gc
    checkPackAllRevisions = checkPackWithMultiDatabaseReferences


class ZRPCConnectionTests(ZEO.tests.ConnectionTests.CommonSetupTearDown):

    def getConfig(self, path, create, read_only):
        return """<mappingstorage 1/>"""

    def checkCatastrophicClientLoopFailure(self):
        # Test what happens when the client loop falls over
        self._storage = self.openClientStorage()

        import zope.testing.loggingsupport
        handler = zope.testing.loggingsupport.InstalledHandler(
            'ZEO.asyncio.client')

        # We no longer implement the event loop, we no longer know
        # how to break it.  We'll just stop it instead for now.
        self._storage._server.loop.call_soon_threadsafe(
            self._storage._server.loop.stop)
        # We wait for the client thread to stop (to avoid a race condition)
        self._storage._server.thread.join(1)
        log = str(handler)
        handler.uninstall()
        self.assertTrue("Client loop stopped unexpectedly" in log)
        self.assertRaises(ClientDisconnected, self._storage.ping)

    def checkExceptionLogsAtError(self):
        # Test the exceptions are logged at error
        self._storage = self.openClientStorage()
        self._dostore(z64, data=MinPO("X" * (10 * 128 * 1024)))

        from zope.testing.loggingsupport import InstalledHandler
        handler = InstalledHandler('ZEO.asyncio.client')
        import ZODB.POSException
        self.assertRaises(TypeError, self._storage.history, z64, None)
        self.assertTrue(re.search(" from server: .*TypeError", str(handler)))

        # POSKeyErrors and ConflictErrors aren't logged:
        handler.clear()
        self.assertRaises(ZODB.POSException.POSKeyError,
                          self._storage.history, None, None)
        handler.uninstall()
        self.assertEqual(str(handler), '')

    def checkConnectionInvalidationOnReconnect(self):

        storage = ClientStorage(self.addr, min_disconnect_poll=0.1)
        self._storage = storage
        assert storage.is_connected()

        class DummyDB:
            _invalidatedCache = 0

            def invalidateCache(self):
                self._invalidatedCache += 1

            def invalidate(*a, **k):
                pass

            transform_record_data = untransform_record_data = \
                lambda self, data: data

        db = DummyDB()
        storage.registerDB(db)

        base = db._invalidatedCache

        # Now we'll force a disconnection and reconnection
        storage._server.loop.call_soon_threadsafe(
            storage._server.client.protocol.connection_lost,
            ValueError('test'))

        # and we'll wait for the storage to be reconnected:
        for i in range(100):
            if storage.is_connected():
                if db._invalidatedCache > base:
                    break
            time.sleep(0.1)
        else:
            raise AssertionError("Couldn't connect to server")

        # Now, the root object in the connection should have been invalidated:
        self.assertEqual(db._invalidatedCache, base+1)


class CommonBlobTests:

    def getConfig(self):
        return """
        <blobstorage 1>
          blob-dir blobs
          <filestorage 2>
            path Data.fs
          </filestorage>
        </blobstorage>
        """

    blobdir = 'blobs'
    blob_cache_dir = 'blob_cache'

    def checkStoreBlob(self):
        from ZODB.blob import Blob
        from ZODB.tests.StorageTestBase import ZERO
        from ZODB.tests.StorageTestBase import zodb_pickle

        somedata = b'a' * 10

        blob = Blob()
        with blob.open('w') as bd_fh:
            bd_fh.write(somedata)
        tfname = bd_fh.name
        oid = self._storage.new_oid()
        data = zodb_pickle(blob)
        self.assertTrue(os.path.exists(tfname))

        t = TransactionMetaData()
        try:
            self._storage.tpc_begin(t)
            self._storage.storeBlob(oid, ZERO, data, tfname, '', t)
            self._storage.tpc_vote(t)
            revid = self._storage.tpc_finish(t)
        except:  # NOQA: E722 bare except
            self._storage.tpc_abort(t)
            raise
        self.assertTrue(not os.path.exists(tfname))
        filename = self._storage.fshelper.getBlobFilename(oid, revid)
        self.assertTrue(os.path.exists(filename))
        with open(filename, 'rb') as f:
            self.assertEqual(somedata, f.read())

    def checkStoreBlob_wrong_partition(self):
        os_rename = os.rename
        try:
            def fail(*a):
                raise OSError
            os.rename = fail
            self.checkStoreBlob()
        finally:
            os.rename = os_rename

    def checkLoadBlob(self):
        from ZODB.blob import Blob
        from ZODB.tests.StorageTestBase import ZERO
        from ZODB.tests.StorageTestBase import zodb_pickle

        somedata = b'a' * 10

        blob = Blob()
        with blob.open('w') as bd_fh:
            bd_fh.write(somedata)
        tfname = bd_fh.name
        oid = self._storage.new_oid()
        data = zodb_pickle(blob)

        t = TransactionMetaData()
        try:
            self._storage.tpc_begin(t)
            self._storage.storeBlob(oid, ZERO, data, tfname, '', t)
            self._storage.tpc_vote(t)
            serial = self._storage.tpc_finish(t)
        except:  # NOQA: E722 bare except
            self._storage.tpc_abort(t)
            raise

        filename = self._storage.loadBlob(oid, serial)
        with open(filename, 'rb') as f:
            self.assertEqual(somedata, f.read())
        self.assertTrue(not (os.stat(filename).st_mode & stat.S_IWRITE))
        self.assertTrue(os.stat(filename).st_mode & stat.S_IREAD)

    def checkTemporaryDirectory(self):
        self.assertEqual(os.path.join(self.blob_cache_dir, 'tmp'),
                         self._storage.temporaryDirectory())

    def checkTransactionBufferCleanup(self):
        oid = self._storage.new_oid()
        with open('blob_file', 'wb') as f:
            f.write(b'I am a happy blob.')
        t = TransactionMetaData()
        self._storage.tpc_begin(t)
        self._storage.storeBlob(
          oid, ZODB.utils.z64, 'foo', 'blob_file', '', t)
        self._storage.tpc_abort(t)
        self._storage.close()


class BlobAdaptedFileStorageTests(FullGenericTests, CommonBlobTests):
    """ZEO backed by a BlobStorage-adapted FileStorage."""

    def checkStoreAndLoadBlob(self):
        from ZODB.blob import Blob
        from ZODB.tests.StorageTestBase import ZERO
        from ZODB.tests.StorageTestBase import zodb_pickle

        somedata_path = os.path.join(self.blob_cache_dir, 'somedata')
        with open(somedata_path, 'w+b') as somedata:
            for i in range(1000000):
                somedata.write(("%s\n" % i).encode('ascii'))

            def check_data(path):
                self.assertTrue(os.path.exists(path))
                somedata.seek(0)
                d1 = d2 = 1
                with open(path, 'rb') as f:
                    while d1 or d2:
                        d1 = f.read(8096)
                        d2 = somedata.read(8096)
                        self.assertEqual(d1, d2)
            somedata.seek(0)

            blob = Blob()
            with blob.open('w') as bd_fh:
                ZODB.utils.cp(somedata, bd_fh)
                bd_fh.close()
                tfname = bd_fh.name
            oid = self._storage.new_oid()
            data = zodb_pickle(blob)
            self.assertTrue(os.path.exists(tfname))

            t = TransactionMetaData()
            try:
                self._storage.tpc_begin(t)
                self._storage.storeBlob(oid, ZERO, data, tfname, '', t)
                self._storage.tpc_vote(t)
                revid = self._storage.tpc_finish(t)
            except:  # NOQA: E722 bare except
                self._storage.tpc_abort(t)
                raise

            # The uncommitted data file should have been removed
            self.assertTrue(not os.path.exists(tfname))

            # The file should be in the cache ...
            filename = self._storage.fshelper.getBlobFilename(oid, revid)
            check_data(filename)

            # ... and on the server
            server_filename = os.path.join(
                self.blobdir,
                ZODB.blob.BushyLayout().getBlobFilePath(oid, revid),
                )

            self.assertTrue(server_filename.startswith(self.blobdir))
            check_data(server_filename)

            # If we remove it from the cache and call loadBlob, it should
            # come back. We can do this in many threads.

            ZODB.blob.remove_committed(filename)
            returns = []
            threads = [
                threading.Thread(
                    target=lambda:
                        returns.append(self._storage.loadBlob(oid, revid))
                    ) for i in range(10)]
            [thread.start() for thread in threads]
            [thread.join() for thread in threads]
            [self.assertEqual(r, filename) for r in returns]
            check_data(filename)


class BlobWritableCacheTests(FullGenericTests, CommonBlobTests):

    blob_cache_dir = 'blobs'
    shared_blob_dir = True


class FauxConn:
    addr = 'x'
    protocol_version = ZEO.asyncio.server.best_protocol_version
    peer_protocol_version = protocol_version

    def async_(self, method, *args):
        pass

    call_soon_threadsafe = async_threadsafe = async_


class StorageServerWrapper:

    def __init__(self, server, storage_id):
        self.storage_id = storage_id
        self.server = ZEO.StorageServer.ZEOStorage(server, server.read_only)
        self.server.notify_connected(FauxConn())
        self.server.register(storage_id, False)

    def sortKey(self):
        return self.storage_id

    def __getattr__(self, name):
        return getattr(self.server, name)

    def registerDB(self, *args):
        pass

    def supportsUndo(self):
        return False

    def new_oid(self):
        return self.server.new_oids(1)[0]

    def tpc_begin(self, transaction):
        self.server.tpc_begin(id(transaction), '', '', {}, None, ' ')

    def tpc_vote(self, transaction):
        return self.server.vote(id(transaction))

    def store(self, oid, serial, data, version_ignored, transaction):
        self.server.storea(oid, serial, data, id(transaction))

    def send_reply(self, _, result):        # Masquerade as conn
        self._result = result

    def tpc_abort(self, transaction):
        self.server.tpc_abort(id(transaction))

    def tpc_finish(self, transaction, func=lambda: None):
        self.server.tpc_finish(id(transaction)).set_sender(0, self)
        return self._result


def multiple_storages_invalidation_queue_is_not_insane():
    """
    >>> from ZEO.StorageServer import StorageServer
    >>> from ZODB.FileStorage import FileStorage
    >>> from ZODB.DB import DB
    >>> from persistent.mapping import PersistentMapping
    >>> from transaction import commit
    >>> fs1 = FileStorage('t1.fs')
    >>> fs2 = FileStorage('t2.fs')
    >>> server = StorageServer(None, storages=dict(fs1=fs1, fs2=fs2))

    >>> s1 = StorageServerWrapper(server, 'fs1')
    >>> s2 = StorageServerWrapper(server, 'fs2')

    >>> db1 = DB(s1); conn1 = db1.open()
    >>> db2 = DB(s2); conn2 = db2.open()

    >>> commit()
    >>> o1 = conn1.root()
    >>> for i in range(10):
    ...     o1.x = PersistentMapping(); o1 = o1.x
    ...     commit()

    >>> last = fs1.lastTransaction()
    >>> for i in range(5):
    ...     o1.x = PersistentMapping(); o1 = o1.x
    ...     commit()

    >>> o2 = conn2.root()
    >>> for i in range(20):
    ...     o2.x = PersistentMapping(); o2 = o2.x
    ...     commit()

    >>> trans, oids = s1.getInvalidations(last)
    >>> from ZODB.utils import u64
    >>> sorted([int(u64(oid)) for oid in oids])
    [10, 11, 12, 13, 14, 15]

    >>> fs1.close(); fs2.close()
    """


def getInvalidationsAfterServerRestart():
    """

Clients were often forced to verify their caches after a server
restart even if there weren't many transactions between the server
restart and the client connect.

Let's create a file storage and stuff some data into it:

    >>> from ZEO.StorageServer import StorageServer, ZEOStorage
    >>> from ZODB.FileStorage import FileStorage
    >>> from ZODB.DB import DB
    >>> from persistent.mapping import PersistentMapping
    >>> fs = FileStorage('t.fs')
    >>> db = DB(fs)
    >>> conn = db.open()
    >>> from transaction import commit
    >>> last = []
    >>> for i in range(100):
    ...     conn.root()[i] = PersistentMapping()
    ...     commit()
    ...     last.append(fs.lastTransaction())
    >>> db.close()

Now we'll open a storage server on the data, simulating a restart:

    >>> fs = FileStorage('t.fs')
    >>> sv = StorageServer(None, dict(fs=fs))
    >>> s = ZEOStorage(sv, sv.read_only)
    >>> s.notify_connected(FauxConn())
    >>> s.register('fs', False) == fs.lastTransaction()
    True

If we ask for the last transaction, we should get the last transaction
we saved:

    >>> s.lastTransaction() == last[-1]
    True

If a storage implements the method lastInvalidations, as FileStorage
does, then the storage server will populate its invalidation data
structure using lastTransactions.

    >>> tid, oids = s.getInvalidations(last[-10])
    >>> tid == last[-1]
    True

    >>> from ZODB.utils import u64
    >>> sorted([int(u64(oid)) for oid in oids])
    [0, 92, 93, 94, 95, 96, 97, 98, 99, 100]

    >>> fs.close()

If a storage doesn't implement lastInvalidations, a client can still
avoid verifying its cache if it was up to date when the server
restarted.  To illustrate this, we'll create a subclass of FileStorage
without this method:

    >>> class FS(FileStorage):
    ...     lastInvalidations = property()

    >>> fs = FS('t.fs')
    >>> sv = StorageServer(None, dict(fs=fs))
    >>> st = StorageServerWrapper(sv, 'fs')
    >>> s = st.server

Now, if we ask for the invalidations since the last committed
transaction, we'll get a result:

    >>> tid, oids = s.getInvalidations(last[-1])
    >>> tid == last[-1]
    True
    >>> oids
    []

    >>> db = DB(st); conn = db.open()
    >>> ob = conn.root()
    >>> for i in range(5):
    ...     ob.x = PersistentMapping(); ob = ob.x
    ...     commit()
    ...     last.append(fs.lastTransaction())

    >>> ntid, oids = s.getInvalidations(tid)
    >>> ntid == last[-1]
    True

    >>> sorted([int(u64(oid)) for oid in oids])
    [0, 101, 102, 103, 104, 105]

Note that in all cases invalidations include both modified objects and objects
that were only created.

    >>> fs.close()
    """


def tpc_finish_error():
    r"""Server errors in tpc_finish weren't handled properly.

    If there are errors applying changes to the client cache, don't
    leave the cache in an inconsistent state.

    >>> addr, admin = start_server()  # NOQA: F821 undefined

    >>> client = ZEO.client(addr)
    >>> db = ZODB.DB(client)
    >>> conn = db.open()
    >>> conn.root.x = 1
    >>> t = conn.transaction_manager.get()
    >>> conn.tpc_begin(t)
    >>> conn.commit(t)
    >>> transaction_meta_data = t.data(conn)
    >>> _ = client.tpc_vote(transaction_meta_data)

    Cause some breakage by messing with the clients transaction
    buffer, sadly, using implementation details:

    >>> tbuf = client._check_trans(transaction_meta_data, 'test')
    >>> tbuf.client_resolved = None

    tpc_finish will fail:

    >>> client.tpc_finish(transaction_meta_data) # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    AttributeError: ...

    >>> client.tpc_abort(transaction_meta_data)
    >>> t.abort()

    But we can still load the saved data:

    >>> conn2 = db.open()
    >>> conn2.root.x
    1

    And we can save new data:

    >>> conn2.root.x += 1
    >>> conn2.transaction_manager.commit()

    >>> db.close()

    >>> stop_server(admin)  # NOQA: F821 undefined
    """


def test_prefetch(self):
    """The client storage prefetch method pre-fetches from the server

    >>> count = 999

    >>> import ZEO
    >>> addr, stop = start_server()  # NOQA: F821 undefined
    >>> conn = ZEO.connection(addr)
    >>> root = conn.root()
    >>> cls = root.__class__
    >>> for i in range(count):
    ...     root[i] = cls()
    >>> conn.transaction_manager.commit()
    >>> oids = [root[i]._p_oid for i in range(count)]
    >>> conn.close()
    >>> conn = ZEO.connection(addr)
    >>> storage = conn.db().storage
    >>> len(storage._cache) <= 1
    True
    >>> storage.prefetch(oids, conn._storage._start)

    The prefetch returns before the cache is filled:

    >>> len(storage._cache) < count
    True

    The ``prefetch`` returns a future.
    As long as it is not yet done, it is part of a reference
    cycle and therefore not immediately garbage collected.
    But a garbage collection run might destroy it.
    Verify that ``prefetch`` nevertheless works correctly.
    >>> import gc
    >>> _ = gc.collect()

    Verify that the cache is filled eventually:

    >>> from zope.testing.wait import wait
    >>> wait((lambda: len(storage._cache) > count), 2)

    >>> loads = storage.server_status()['loads']

    Now if we reload the data, it will be satisfied from the cache:

    >>> for oid in oids:
    ...     _ = conn._storage.load(oid)

    >>> storage.server_status()['loads'] == loads
    True

    >>> conn.close()
    """


def client_has_newer_data_than_server():
    """It is bad if a client has newer data than the server.

    >>> db = ZODB.DB('Data.fs')
    >>> db.close()
    >>> r = shutil.copyfile('Data.fs', 'Data.save')
    >>> addr, admin = start_server(keep=1)  # NOQA: F821 undefined
    >>> db = ZEO.DB(addr, name='client')
    >>> wait_connected(db.storage)  # NOQA: F821 undefined
    >>> conn = db.open()
    >>> conn.root().x = 1
    >>> transaction.commit()

    OK, we've added some data to the storage and the client cache has
    the new data. Now, we'll stop the server, put back the old data, and
    see what happens. :)

    >>> stop_server(admin)  # NOQA: F821 undefined
    >>> r = shutil.copyfile('Data.save', 'Data.fs')

    >>> import zope.testing.loggingsupport
    >>> handler = zope.testing.loggingsupport.InstalledHandler(
    ...     'ZEO', level=logging.ERROR)
    >>> formatter = logging.Formatter('%(name)s %(levelname)s %(message)s')

    >>> _, admin = start_server(addr=addr)  # NOQA: F821 undefined

    >>> wait_until('got enough errors', lambda:  # NOQA: F821 undefined
    ...    len([x for x in handler.records
    ...         if x.levelname == 'CRITICAL' and
    ...            'Client cache is out of sync with the server.' in x.msg
    ...         ]) >= 2, 30)

    Note that the errors repeat because the client keeps on trying to connect.
    We have to wait rather long as the client waits about 10 s
    before a retrial.

    >>> db.close()
    >>> handler.uninstall()
    >>> stop_server(admin)  # NOQA: F821 undefined

    """


def history_over_zeo():
    """
    >>> addr, _ = start_server()  # NOQA: F821 undefined
    >>> db = ZEO.DB(addr)
    >>> wait_connected(db.storage)  # NOQA: F821 undefined
    >>> conn = db.open()
    >>> conn.root().x = 0
    >>> transaction.commit()
    >>> len(db.history(conn.root()._p_oid, 99))
    2

    >>> db.close()
    """


def dont_log_poskeyerrors_on_server():
    """
    >>> addr, admin = start_server(log='server.log')  # NOQA: F821 undefined
    >>> cs = ClientStorage(addr)
    >>> cs.load(ZODB.utils.p64(1))
    Traceback (most recent call last):
    ...
    POSKeyError: 0x01

    >>> cs.close()
    >>> stop_server(admin)  # NOQA: F821 undefined
    >>> with open('server.log') as f:
    ...     'POSKeyError' in f.read()
    False
    """


def open_convenience():
    """Often, we just want to open a single connection.

    >>> addr, _ = start_server(path='data.fs')  # NOQA: F821 undefined
    >>> conn = ZEO.connection(addr)
    >>> conn.root()
    {}

    >>> conn.root()['x'] = 1
    >>> transaction.commit()
    >>> conn.close()

    Let's make sure the database was cloased when we closed the
    connection, and that the data is there.

    >>> db = ZEO.DB(addr)
    >>> conn = db.open()
    >>> conn.root()
    {'x': 1}
    >>> db.close()
    """


def client_asyncore_thread_has_name():
    """
    >>> addr, _ = start_server()  # NOQA: F821 undefined
    >>> db = ZEO.DB(addr)
    >>> any(t for t in threading.enumerate()
    ...     if ' zeo client networking thread' in t.name)
    True
    >>> db.close()
    """


def runzeo_without_configfile():
    r"""
    >>> with open('runzeo', 'w') as r:
    ...     _ = r.write('''
    ... import sys
    ... sys.path[:] = %r
    ... import ZEO.runzeo
    ... ZEO.runzeo.main(sys.argv[1:])
    ... ''' % sys.path)

    >>> import subprocess, re
    >>> proc = subprocess.Popen(
    ...     [sys.executable, 'runzeo', '-a:0', '-ft', '--test'],
    ...     stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
    ...     )
    >>> proc.wait()
    0

    # Note: the heuristic to determine the "instance home" can fail
    # causing an initial error message in the process output.
    # We check for this and remove it in this case.
    >>> output = re.sub(br'\d\d+|[:]', b'', proc.stdout.read()).decode('ascii')
    >>> if "ERROR" in output.splitlines()[1]:
    ...    output = "\n".join(output.splitlines()[2:])
    >>> print(output) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    ------
    --T INFO ZEO.runzeo () opening storage '1' using FileStorage
    ------
    --T INFO ZEO.StorageServer StorageServer created RW with storages 1RWt...
    ------
    --T INFO ZEO.asyncio... listening on ...
    testing exit immediately
    ------
    --T INFO ZEO.StorageServer closing storage '1'...

    >>> proc.stdout.close()
    """


def close_client_storage_w_invalidations():
    r"""
Invalidations could cause errors when closing client storages,

    >>> addr, _ = start_server()  # NOQA: F821 undefined
    >>> writing = threading.Event()
    >>> def mad_write_thread():
    ...     global writing
    ...     conn = ZEO.connection(addr)
    ...     writing.set()
    ...     while writing.is_set():
    ...         conn.root.x = 1
    ...         transaction.commit()
    ...     conn.close()

    >>> thread = threading.Thread(target=mad_write_thread)
    >>> thread.daemon = True
    >>> thread.start()
    >>> _ = writing.wait()
    >>> time.sleep(.01)
    >>> for i in range(10):
    ...     conn = ZEO.connection(addr)
    ...     _ = conn._storage.load(b'\0'*8)
    ...     conn.close()

    >>> writing.clear()
    >>> thread.join(1)
    """


def convenient_to_pass_port_to_client_and_ZEO_dot_client():
    """Jim hates typing

    >>> addr, _ = start_server()  # NOQA: F821 undefined
    >>> client = ZEO.client(addr[1])
    >>> client.__name__ == "('127.0.0.1', %s)" % addr[1]
    True

    >>> client.close()
    """


def test_server_status():
    """
    You can get server status using the server_status method.

    >>> addr, _ = start_server(  # NOQA: F821 undefined
    ...             zeo_conf=dict(transaction_timeout=1))
    >>> db = ZEO.DB(addr)
    >>> pprint.pprint(db.storage.server_status(), width=40)
    {'aborts': 0,
     'active_txns': 0,
     'commits': 1,
     'conflicts': 0,
     'conflicts_resolved': 0,
     'connections': 1,
     'last-transaction': '03ac11b771fa1c00',
     'loads': 1,
     'lock_time': None,
     'start': 'Tue May  4 10:55:20 2010',
     'stores': 1,
     'timeout-thread-is-alive': True,
     'waiting': 0}

    >>> db.close()
    """


def test_ruok():
    """
    You can also get server status using the ruok protocol.

    >>> addr, _ = start_server(  # NOQA: F821 undefined
    ...             zeo_conf=dict(transaction_timeout=1))
    >>> db = ZEO.DB(addr) # force a transaction :)
    >>> import json, socket, struct
    >>> s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    >>> s.connect(addr)
    >>> writer = s.makefile(mode='wb')
    >>> _ = writer.write(struct.pack(">I", 4)+b"ruok")
    >>> writer.close()
    >>> proto = s.recv(struct.unpack(">I", s.recv(4))[0])
    >>> data = json.loads(
    ...     s.recv(struct.unpack(">I", s.recv(4))[0]).decode("ascii"))
    >>> pprint.pprint(data['1'])
    {'aborts': 0,
     'active_txns': 0,
     'commits': 1,
     'conflicts': 0,
     'conflicts_resolved': 0,
     'connections': 1,
     'last-transaction': '03ac11cd11372499',
     'loads': 1,
     'lock_time': None,
     'start': 'Sun Jan  4 09:37:03 2015',
     'stores': 1,
     'timeout-thread-is-alive': True,
     'waiting': 0}
    >>> db.close(); s.close()
    """


def client_labels():
    """
When looking at server logs, for servers with lots of clients coming
from the same machine, it can be very difficult to correlate server
log entries with actual clients.  It's possible, sort of, but tedious.

You can make this easier by passing a label to the ClientStorage
constructor.

    >>> addr, _ = start_server(log='server.log')  # NOQA: F821 undefined
    >>> db = ZEO.DB(addr, client_label='test-label-1')
    >>> db.close()
    >>> @wait_until  # NOQA: F821 undefined
    ... def check_for_test_label_1():
    ...    with open('server.log') as f:
    ...        for line in f:
    ...            if 'test-label-1' in line:
    ...                print(line.split()[1:4])
    ...                return True
    ['INFO', 'ZEO.StorageServer', '(test-label-1']

You can specify the client label via a configuration file as well:

    >>> import ZODB.config
    >>> db = ZODB.config.databaseFromString('''
    ... <zodb>
    ...    <zeoclient>
    ...       server :%s
    ...       client-label test-label-2
    ...    </zeoclient>
    ... </zodb>
    ... ''' % addr[1])
    >>> db.close()
    >>> @wait_until  # NOQA: F821 undefined
    ... def check_for_test_label_2():
    ...     with open('server.log') as f:
    ...         for line in f:
    ...             if 'test-label-2' in line:
    ...                 print(line.split()[1:4])
    ...                 return True
    ['INFO', 'ZEO.StorageServer', '(test-label-2']

    """


def invalidate_client_cache_entry_on_server_commit_error():
    """

When the serials returned during commit includes an error, typically a
conflict error, invalidate the cache entry.  This is important when
the cache is messed up.

    >>> addr, _ = start_server()  # NOQA: F821 undefined
    >>> conn1 = ZEO.connection(addr)
    >>> conn1.root.x = conn1.root().__class__()
    >>> transaction.commit()
    >>> conn1.root.x
    {}

    >>> cs = ZEO.ClientStorage.ClientStorage(addr, client='cache')
    >>> conn2 = ZODB.connection(cs)
    >>> conn2.root.x
    {}

    >>> conn2.close()
    >>> cs.close()

    >>> conn1.root.x['x'] = 1
    >>> transaction.commit()
    >>> conn1.root.x
    {'x': 1}

Now, let's screw up the cache by making it have a last tid that is later than
the root serial.

    >>> import ZEO.cache
    >>> cache = ZEO.cache.ClientCache('cache-1.zec')
    >>> cache.setLastTid(p64(u64(conn1.root.x._p_serial)+1))
    >>> cache.close()

We'll also update the server so that it's last tid is newer than the cache's:

    >>> conn1.root.y = 1
    >>> transaction.commit()
    >>> conn1.root.y = 2
    >>> transaction.commit()

Now, if we reopen the client storage, we'll get the wrong root:

    >>> cs = ZEO.ClientStorage.ClientStorage(addr, client='cache')
    >>> conn2 = ZODB.connection(cs)
    >>> conn2.root.x
    {}

And, we'll get a conflict error if we try to modify it:

    >>> conn2.root.x['y'] = 1
    >>> transaction.commit() # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    ConflictError: ...

But, if we abort, we'll get up to date data and we'll see the changes.

    >>> transaction.abort()
    >>> conn2.root.x
    {'x': 1}
    >>> conn2.root.x['y'] = 1
    >>> transaction.commit()
    >>> sorted(conn2.root.x.items())
    [('x', 1), ('y', 1)]

    >>> conn2.close()
    >>> cs.close()
    >>> conn1.close()
    """


script_template = """
import sys
sys.path[:] = %(path)r

%(src)s

"""


def generate_script(name, src):
    with open(name, 'w') as f:
        f.write(script_template % dict(
            exe=sys.executable,
            path=sys.path,
            src=src,
        ))


def read(filename):
    with open(filename) as f:
        return f.read()


def runzeo_logrotate_on_sigusr2():
    """
    >>> from ZEO.tests.forker import get_port
    >>> port = get_port()
    >>> with open('c', 'w') as r:
    ...    _ = r.write('''
    ... <zeo>
    ...    address %s
    ... </zeo>
    ... <mappingstorage>
    ... </mappingstorage>
    ... <eventlog>
    ...    <logfile>
    ...       path l
    ...    </logfile>
    ... </eventlog>
    ... ''' % port)
    >>> generate_script('s', '''
    ... import ZEO.runzeo
    ... ZEO.runzeo.main()
    ... ''')
    >>> import subprocess
    >>> p = subprocess.Popen([sys.executable, 's', '-Cc'], close_fds=True)
    >>> wait_until('started',  # NOQA: F821 undefined
    ...       lambda: os.path.exists('l') and ('listening on' in read('l'))
    ...     )

    >>> oldlog = read('l')
    >>> os.rename('l', 'o')
    >>> os.kill(p.pid, signal.SIGUSR2)

    >>> s = ClientStorage(port)
    >>> s.close()
    >>> wait_until('See logging',  # NOQA: F821 undefined
    ...            lambda: ('Log files ' in read('l')))
    >>> read('o') == oldlog  # No new data in old log
    True

    # Cleanup:

    >>> os.kill(p.pid, signal.SIGKILL)
    >>> _ = p.wait()
    """


def unix_domain_sockets():
    """Make sure unix domain sockets work

    >>> addr, _ = start_server(port='./sock')  # NOQA: F821 undefined

    >>> c = ZEO.connection(addr)
    >>> c.root.x = 1
    >>> transaction.commit()
    >>> c.close()
    """


def gracefully_handle_abort_while_storing_many_blobs():
    r"""

    >>> import logging, sys
    >>> old_level = logging.getLogger().getEffectiveLevel()
    >>> logging.getLogger().setLevel(logging.ERROR)
    >>> handler = logging.StreamHandler(sys.stdout)
    >>> logging.getLogger().addHandler(handler)

    >>> addr, _ = start_server(blob_dir='blobs')  # NOQA: F821 undefined
    >>> client = ZEO.client(addr, blob_dir='cblobs')
    >>> c = ZODB.connection(client)
    >>> c.root.x = ZODB.blob.Blob(b'z'*(1<<20))
    >>> c.root.y = ZODB.blob.Blob(b'z'*(1<<2))
    >>> t = c.transaction_manager.get()
    >>> c.tpc_begin(t)
    >>> c.commit(t)

We've called commit, but the blob sends are queued.  We'll call abort
right away, which will delete the temporary blob files.  The queued
iterators will try to open these files.

    >>> c.tpc_abort(t)

Now we'll try to use the connection, mainly to wait for everything to
get processed. Before we fixed this by making tpc_finish a synchronous
call to the server. we'd get some sort of error here.

    >>> _ = client._call('loadBefore', b'\0'*8, maxtid)

    >>> c.close()

    >>> logging.getLogger().removeHandler(handler)
    >>> logging.getLogger().setLevel(old_level)



    """


def ClientDisconnected_errors_are_TransientErrors():
    """
    >>> from ZEO.Exceptions import ClientDisconnected
    >>> from transaction.interfaces import TransientError
    >>> issubclass(ClientDisconnected, TransientError)
    True
    """


if os.environ.get('ZEO_MSGPACK'):
    def test_runzeo_msgpack_support():
        """
        >>> import ZEO

        >>> a, s = ZEO.server(threaded=False)
        >>> conn = ZEO.connection(a)
        >>> str(conn.db().storage.protocol_version.decode('ascii'))
        'M5'
        >>> conn.close(); s()
        """
else:
    def test_runzeo_msgpack_support():
        """
        >>> import ZEO

        >>> a, s = ZEO.server(threaded=False)
        >>> conn = ZEO.connection(a)
        >>> str(conn.db().storage.protocol_version.decode('ascii'))
        'Z5'
        >>> conn.close(); s()

        >>> a, s = ZEO.server(zeo_conf=dict(msgpack=True), threaded=False)
        >>> conn = ZEO.connection(a)
        >>> str(conn.db().storage.protocol_version.decode('ascii'))
        'M5'
        >>> conn.close(); s()
        """

if WIN:
    del runzeo_logrotate_on_sigusr2
    del unix_domain_sockets


def work_with_multiprocessing_process(name, addr, q):
    conn = ZEO.connection(addr)
    q.put((name, conn.root.x))
    conn.close()


class MultiprocessingTests(unittest.TestCase):

    layer = ZODB.tests.util.MininalTestLayer('work_with_multiprocessing')

    def test_work_with_multiprocessing(self):
        "Client storage should work with multi-processing."

        # Gaaa, zope.testing.runner.FakeInputContinueGenerator has no close
        if not hasattr(sys.stdin, 'close'):
            sys.stdin.close = lambda: None
        if not hasattr(sys.stdin, 'fileno'):
            sys.stdin.fileno = lambda: -1

        self.globs = {}
        forker.setUp(self)
        addr, adminaddr = self.globs['start_server']()
        conn = ZEO.connection(addr)
        conn.root.x = 1
        transaction.commit()
        q = multiprocessing.Queue()
        processes = [multiprocessing.Process(
            target=work_with_multiprocessing_process,
            args=(i, addr, q))
                        for i in range(3)]
        _ = [p.start() for p in processes]
        self.assertEqual(sorted(q.get(timeout=300) for p in processes),
                         [(0, 1), (1, 1), (2, 1)])

        _ = [p.join(30) for p in processes]
        conn.close()
        zope.testing.setupstack.tearDown(self)


def quick_close_doesnt_kill_server():
    r"""

    Start a server:

    >>> from .testssl import server_config, client_ssl
    >>> addr, _ = start_server(zeo_conf=server_config)  # NOQA: F821 undefined

    Now connect and immediately disconnect. This caused the server to
    die in the past:

    >>> import socket, struct
    >>> for i in range(5):
    ...     s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    ...     s.setsockopt(socket.SOL_SOCKET, socket.SO_LINGER,
    ...                  struct.pack('ii', 1, 0))
    ...     s.connect(addr)
    ...     s.close()


    >>> print("\n\nXXX WARNING: running quick_close_doesnt_kill_server "
    ...       "with ssl as hack pending http://bugs.python.org/issue27386\n",
    ...       file=sys.stderr)  # Intentional long line to be annoying
    ...                         # until this is fixed

    Now we should be able to connect as normal:

    >>> db = ZEO.DB(addr, ssl=client_ssl())
    >>> db.storage.is_connected()
    True

    >>> db.close()
    """


def can_use_empty_string_for_local_host_on_client():
    """We should be able to spell localhost with ''.

    >>> (_, port), _ = start_server()  # NOQA: F821 undefined name
    >>> conn = ZEO.connection(('', port))
    >>> conn.root()
    {}
    >>> conn.root.x = 1
    >>> transaction.commit()

    >>> conn.close()
    """


slow_test_classes = [
    BlobAdaptedFileStorageTests, BlobWritableCacheTests,
    # keep to test a storage without blobs
    # and with in memory store (may have different latency than
    # ``FileStorage`` and therefore expose other race conditions)
    MappingStorageTests,
    # drop to save time
    # DemoStorageTests,
    # FileStorageTests,
    # FileStorageHexTests, FileStorageClientHexTests,
    FileStorageLoadDelayedTests,
    FileStorageSSLTests,
    ]

quick_test_classes = [FileStorageRecoveryTests, ZRPCConnectionTests]


class ServerManagingClientStorage(ClientStorage):

    def __init__(self, name, blob_dir, shared=False, extrafsoptions=''):
        if shared:
            server_blob_dir = blob_dir
        else:
            server_blob_dir = 'server-'+blob_dir
        self.globs = {}
        addr, stop = forker.start_zeo_server(
            f"""
            <blobstorage>
                blob-dir {server_blob_dir}
                <filestorage>
                   path {name}.fs
                   {extrafsoptions}
                </filestorage>
            </blobstorage>
            """
            )
        zope.testing.setupstack.register(self, stop)
        if shared:
            ClientStorage.__init__(self, addr, blob_dir=blob_dir,
                                   shared_blob_dir=True)
        else:
            ClientStorage.__init__(self, addr, blob_dir=blob_dir)

    def close(self):
        ClientStorage.close(self)
        zope.testing.setupstack.tearDown(self)


def create_storage_shared(name, blob_dir):
    return ServerManagingClientStorage(name, blob_dir, True)


class ServerManagingClientStorageForIExternalGCTest(
        ServerManagingClientStorage):

    def pack(self, t=None, referencesf=None):
        ServerManagingClientStorage.pack(self, t, referencesf, wait=True)
        # Packing doesn't clear old versions out of zeo client caches,
        # so we'll clear the caches.
        self._cache.clear()
        ZEO.ClientStorage._check_blob_cache_size(self.blob_dir, 0)


def test_suite():
    suite = unittest.TestSuite((
        unittest.defaultTestLoader.loadTestsFromTestCase(
            Test_convenience_functions),
    ))

    zeo = unittest.TestSuite()
    zeo.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(
        ZODB.tests.util.AAAA_Test_Runner_Hack))
    patterns = [
        (re.compile(r"'start': '[^\n]+'"), 'start'),
        (re.compile(r"'last-transaction': '[0-9a-f]+'"),
         'last-transaction'),
        (re.compile("ZODB.POSException.ConflictError"), "ConflictError"),
        (re.compile("ZODB.POSException.POSKeyError"), "POSKeyError"),
        (re.compile("ZEO.Exceptions.ClientStorageError"),
         "ClientStorageError"),
        (re.compile("ZEO.Exceptions.ClientDisconnected"),
         "ClientDisconnected"),
        (re.compile(r"\[Errno \d+\]"), '[Errno N]'),
        (re.compile(r"loads=\d+\.\d+"), 'loads=42.42'),
        # GHA prints this for PyPy3 to stdout:
        (re.compile(
            r"/home/runner/work/ZEO/ZEO/src/ZEO/tests/server.pem None\n"),
         ''),
        ]
    zeo.addTest(doctest.DocTestSuite(
        setUp=forker.setUp, tearDown=zope.testing.setupstack.tearDown,
        checker=renormalizing.RENormalizing(patterns),
        ))
    zeo.addTest(doctest.DocTestSuite(
            ZEO.tests.IterationTests,
            setUp=forker.setUp, tearDown=zope.testing.setupstack.tearDown,
            checker=renormalizing.RENormalizing((
                    (re.compile("ZEO.Exceptions.ClientDisconnected"),
                     "ClientDisconnected"),
                    )),
            ))

    def add_tests(to, case):
        """add tests from *case* to *to*.

        This adds tests with prefixes ``check`` and ``test``
        to be compatible with ``ZODB<6`` and ``ZODB>=6``.
        """
        for prefix in ('check', 'test'):
            test_loader = unittest.TestLoader()
            test_loader.testMethodPrefix = prefix
            to.addTest(test_loader.loadTestsFromTestCase(case))

    add_tests(zeo, ClientConflictResolutionTests)
    zeo.layer = ZODB.tests.util.MininalTestLayer('testZeo-misc')
    suite.addTest(zeo)

    zeo = unittest.TestSuite()
    zeo.addTest(
        doctest.DocFileSuite(
            'zdoptions.test',
            'drop_cache_rather_than_verify.txt', 'client-config.test',
            'protocols.test', 'zeo_blob_cache.test', 'invalidation-age.txt',
            '../nagios.rst',
            setUp=forker.setUp, tearDown=zope.testing.setupstack.tearDown,
            checker=renormalizing.RENormalizing(patterns),
            ),
        )
    zeo.addTest(PackableStorage.IExternalGC_suite(
        lambda:
        ServerManagingClientStorageForIExternalGCTest(
            'data.fs', 'blobs', extrafsoptions='pack-gc false')
        ))
    for klass in quick_test_classes:
        add_tests(zeo, klass)
    zeo.layer = ZODB.tests.util.MininalTestLayer('testZeo-misc2')
    suite.addTest(zeo)

    # tests that often fail, maybe if they have their own layers
    for name in 'zeo-fan-out.test', 'new_addr.test':
        zeo = unittest.TestSuite()
        zeo.addTest(
            doctest.DocFileSuite(
                name,
                setUp=forker.setUp, tearDown=zope.testing.setupstack.tearDown,
                checker=renormalizing.RENormalizing(patterns),
                ),
            )
        zeo.layer = ZODB.tests.util.MininalTestLayer('testZeo-' + name)
        suite.addTest(zeo)

    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(
        MultiprocessingTests))

    # Put the heavyweights in their own layers
    for klass in slow_test_classes:
        sub = unittest.TestSuite()
        add_tests(sub, klass)
        sub.layer = ZODB.tests.util.MininalTestLayer(klass.__name__)
        suite.addTest(sub)

    suite.addTest(ZODB.tests.testblob.storage_reusable_suite(
        'ClientStorageNonSharedBlobs', ServerManagingClientStorage))
    suite.addTest(ZODB.tests.testblob.storage_reusable_suite(
        'ClientStorageSharedBlobs', create_storage_shared))

    from .threaded import threaded_server_tests
    dynamic_server_ports_suite = doctest.DocFileSuite(
        'dynamic_server_ports.test',
        setUp=forker.setUp, tearDown=zope.testing.setupstack.tearDown,
        checker=renormalizing.RENormalizing(patterns),
        )
    dynamic_server_ports_suite.layer = threaded_server_tests
    suite.addTest(dynamic_server_ports_suite)

    return suite



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""%prog [options] address

Where the address is an IPV6 address of the form: [addr]:port, an IPV4
address of the form: addr:port, or the name of a unix-domain socket file.
"""
import json
import optparse
import os
import re
import socket
import struct
import sys
import time


NO_TRANSACTION = '0'*16

nodiff_names = 'active_txns connections waiting'.split()
diff_names = 'aborts commits conflicts conflicts_resolved loads stores'.split()

per_times = dict(seconds=1.0, minutes=60.0, hours=3600.0, days=86400.0)


def new_metric(metrics, storage_id, name, value):
    if storage_id == '1':
        label = name
    else:
        if ' ' in storage_id:
            label = f"'{storage_id}:{name}'"
        else:
            label = f'{storage_id}:{name}'
    metrics.append(f'{label}={value}')


def result(messages, metrics=(), status=None):
    if metrics:
        messages[0] += '|' + metrics[0]
    if len(metrics) > 1:
        messages.append('| ' + '\n '.join(metrics[1:]))
    print('\n'.join(messages))
    return status


def error(message):
    return result((message, ), (), 2)


def warn(message):
    return result((message, ), (), 1)


def check(addr, output_metrics, status, per):
    m = re.match(r'\[(\S+)\]:(\d+)$', addr)
    if m:
        addr = m.group(1), int(m.group(2))
        s = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
    else:
        m = re.match(r'(\S+):(\d+)$', addr)
        if m:
            addr = m.group(1), int(m.group(2))
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        else:
            s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    try:
        s.connect(addr)
    except OSError as err:
        s.close()
        return error("Can't connect %s" % err)

    s.sendall(b'\x00\x00\x00\x04ruok')
    proto = s.recv(struct.unpack(">I", s.recv(4))[0])  # NOQA: F841 unused
    datas = s.recv(struct.unpack(">I", s.recv(4))[0])
    s.close()
    data = json.loads(datas.decode("ascii"))
    if not data:
        return warn("No storages")

    metrics = []
    messages = []
    level = 0
    if output_metrics:
        for storage_id, sdata in sorted(data.items()):
            for name in nodiff_names:
                new_metric(metrics, storage_id, name, sdata[name])

        if status:
            now = time.time()
            if os.path.exists(status):
                dt = now - os.stat(status).st_mtime
                if dt > 0:  # sanity :)
                    with open(status) as f:  # Read previous
                        old = json.loads(f.read())
                    dt /= per_times[per]
                    for storage_id, sdata in sorted(data.items()):
                        sdata['sameple-time'] = now
                        if storage_id in old:
                            sold = old[storage_id]
                            for name in diff_names:
                                v = (sdata[name] - sold[name]) / dt
                                new_metric(metrics, storage_id, name, v)
            with open(status, 'w') as f:  # save current
                f.write(json.dumps(data))

    for storage_id, sdata in sorted(data.items()):
        if sdata['last-transaction'] == NO_TRANSACTION:
            messages.append("Empty storage %r" % storage_id)
            level = max(level, 1)
    if not messages:
        messages.append('OK')
    return result(messages, metrics, level or None)


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    parser = optparse.OptionParser(__doc__)
    parser.add_option(
        '-m', '--output-metrics', action="store_true",
        help="Output metrics.",
        )
    parser.add_option(
        '-s', '--status-path',
        help="Path to status file, needed to get rate metrics",
        )
    parser.add_option(
        '-u', '--time-units', type='choice', default='minutes',
        choices=['seconds', 'minutes', 'hours', 'days'],
        help="Time unit for rate metrics",
        )
    (options, args) = parser.parse_args(args)
    [addr] = args
    return check(
        addr, options.output_metrics, options.status_path, options.time_units)


if __name__ == '__main__':
    main()



import argparse
import logging
import multiprocessing
import os
import persistent
import re
import time

import ZODB
from BTrees.IOBTree import BTree

import ZEO

#logging.basicConfig(level='DEBUG')

class P(persistent.Persistent):

    children = ()


parser = argparse.ArgumentParser("zodbshootout-inspired performance exercise")

parser.add_argument('--concurrency', '-c', type=int, default=1)
parser.add_argument('--client-hosts', '-C', type=int, default=1)
parser.add_argument('--object-size', '-o', type=int, default=999)
parser.add_argument('--transaction-size', '-t', type=int, default=3)
parser.add_argument('--repetitions', '-r', type=int, default=1000)
parser.add_argument('--demo-storage-baseline', '-d', action='store_true')
parser.add_argument('--address', '-a')
parser.add_argument('--ssl', action='store_true')
parser.add_argument('--server-sync', action='store_true')
parser.add_argument('--profile')
parser.add_argument('--save', '-s', help='save results to file')
parser.add_argument('--name', '-n', help='Run name')
parser.add_argument('--read-only', '-W', action='store_true')

def shoot():
    """zodbshootout-inspired performance exercise.
    """

    options = parser.parse_args()
    concurrency = options.concurrency
    object_size = options.object_size
    transaction_size = options.transaction_size
    repetitions = options.repetitions

    if options.ssl:
        from ZEO.tests.testssl import server_config, client_ssl
    else:
        server_config = None
        client_ssl = lambda : None

    if options.demo_storage_baseline:
        db_factory = None
        headings = ('add', 'update', 'read')
        stop = lambda : None
    else:
        if options.address:
            addr = options.address
            if ':' in addr:
                host, port = addr.split(':')
                addr = host, int(port)
            else:
                addr = '127.0.0.1', int(addr)
            stop = lambda : None
        else:
            if os.path.exists('perf.fs'):
                os.remove('perf.fs')

            try:
                addr, stop = ZEO.server(
                    threaded=False, path='perf.fs', zeo_conf=server_config)
            except TypeError:
                # ZEO 4
                addr, stop = ZEO.server()

        db_factory = lambda : ZEO.DB(
            addr,
            ssl=client_ssl(),
            wait_timeout=9999,
            server_sync=options.server_sync)

        if options.read_only:
            headings = ('read', 'prefetch')
        else:
            headings = ('add', 'update', 'cached', 'read', 'prefetch')
            # Initialize database
            db = db_factory()
            with db.transaction() as conn:
                conn.root.speedtest = speedtest = BTree()
                for ic in range(concurrency):
                    speedtest[ic] = data = BTree()
                    for ir in range(repetitions):
                        data[ir] = P()

            db.pack()
            db.close()


    print('Times per operation in microseconds (o=%s, t=%s, c=%s)' % (
        object_size, transaction_size, concurrency))
    print(' %12s' * 5 % ('op', 'min', 'mean', 'median', 'max'))

    queues = [(multiprocessing.Queue(), multiprocessing.Queue())
              for ip in range(concurrency)]

    if options.save:
        save_file = open(options.save, 'a')
    else:
        save_file = None

    if concurrency > 1 or save_file:
        processes = [
            multiprocessing.Process(
                target=run_test,
                args=(db_factory, ip, queues[ip][0], queues[ip][1],
                      object_size, transaction_size, repetitions,
                      options.read_only),
                )
            for ip in range(concurrency)
            ]

        for p in processes:
            p.daemon = True
            p.start()

        for iqueue, oqueue in queues:
            oqueue.get(timeout=9) # ready?

        for name in headings:
            for iqueue, oqueue in queues:
                iqueue.put(None)
            data = [oqueue.get(timeout=999) / repetitions
                    for _, oqueue in queues]
            summarize(name, data)
            if save_file:
                save_file.write('\t'.join(
                    map(str,
                        (options.name,
                         object_size, transaction_size,
                         concurrency * options.client_hosts, repetitions,
                         options.server_sync, options.ssl,
                         options.demo_storage_baseline,
                         options.address,
                         name, sum(data)/len(data))
                        )
                    ) + '\n')

        for p in processes:
            p.join(1)

    else:
        [(iqueue, oqueue)] = queues
        for name in headings:
            iqueue.put(None)

        if options.profile:
            import cProfile
            profiler = cProfile.Profile()
            profiler.enable()
        else:
            profiler = None
        run_test(db_factory, 0, iqueue, oqueue,
                 object_size, transaction_size, repetitions,
                 options.read_only)
        oqueue.get(timeout=9) # ready?
        if profiler is not None:
            profiler.disable()
            profiler.dump_stats(options.profile)

        for name in headings:
            summarize(name, [oqueue.get(timeout=999) / repetitions])

    stop()

def summarize(name, results):
    results = [int(t*1000000) for t in results]
    l = len(results)
    if l > 1:
        print(' %12s' * 5 % (
            name, min(results), round(sum(results)/l), results[l//2],
            max(results)
            ))
    else:
        print(' %12s' * 2 % (name, results[0]))

def run_test(db_factory, process_index, iqueue, oqueue,
             object_size, transaction_size, repetitions, read_only):

    if db_factory is None:
        # No factory. Compute baseline numbers using DemoStorage
        db = ZODB.DB(None)
        with db.transaction() as conn:
            conn.root.speedtest = speedtest = BTree()
            speedtest[process_index] = data = BTree()
            for ir in range(repetitions):
                data[ir] = P()
    else:
        db = db_factory()

    conn = db.open()
    data = conn.root.speedtest[process_index]
    slots = list(data.values()) # Get data slots loaded
    oqueue.put('ready')

    if not read_only:
        # add test
        iqueue.get()
        start = time.time()
        for slot in slots:
            conn.transaction_manager.begin()
            for it in range(transaction_size):
                p = P()
                p.data = 'x' * object_size
                slot.children += (p, )
            conn.transaction_manager.commit()

        conn.transaction_manager.commit()
        oqueue.put(time.time() - start)

        iqueue.get()
        # update test
        start = time.time()
        for slot in slots:
            conn.transaction_manager.begin()
            for it in range(transaction_size):
                slot.children[it].data = 'y' * object_size
            conn.transaction_manager.commit()
        oqueue.put(time.time() - start)

        iqueue.get()

        # hot, With db cache, but not object cache
        conn.cacheMinimize()
        for slot in slots:
            _ = slot.children
        start = time.time()
        for slot in slots:
            conn.transaction_manager.begin()
            for it in range(transaction_size):
                assert slot.children[it].data == 'y' * object_size
            conn.transaction_manager.commit()
        oqueue.put(time.time() - start)

    if db_factory is None:
        return

    storage = db.storage

    # cold, With no cache
    storage._cache.clear()
    conn.cacheMinimize()
    for slot in slots[:repetitions]:
        _ = slot.children

    iqueue.get()
    start = time.time()
    for slot in slots[:repetitions]:
        conn.transaction_manager.begin()
        for it in range(transaction_size):
            assert slot.children[it].data == 'y' * object_size
        conn.transaction_manager.commit()
    oqueue.put(time.time() - start)

    # cold, With no cache, but with prefetch
    oids = []
    for slot in slots[:repetitions]:
        soids = []
        oids.append(soids)
        for it in range(transaction_size):
            soids.append(slot.children[it]._p_oid)

    storage._cache.clear()
    conn.cacheMinimize()
    for slot in slots[:repetitions]:
        _ = slot.children

    iqueue.get()
    start = time.time()
    for slot, soids in zip(slots[:repetitions], oids):
        conn.transaction_manager.begin()
        storage.prefetch(soids, conn._storage._start)
        for it in range(transaction_size):
            assert slot.children[it].data == 'y' * object_size
        conn.transaction_manager.commit()
    oqueue.put(time.time() - start)

    conn.close()
    db.close()

if __name__ == '__main__':
    shoot()



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""A Thread base class for use with unittest."""
import sys
import threading


class TestThread(threading.Thread):
    """Base class for defining threads that run from unittest.

    The subclass should define a testrun() method instead of a run()
    method.

    Call cleanup() when the test is done with the thread, instead of join().
    If the thread exits with an uncaught exception, it's captured and
    re-raised when cleanup() is called.  cleanup() should be called by
    the main thread!  Trying to tell unittest that a test failed from
    another thread creates a nightmare of timing-depending cascading
    failures and missed errors (tracebacks that show up on the screen,
    but don't cause unittest to believe the test failed).

    cleanup() also joins the thread.  If the thread ended without raising
    an uncaught exception, and the join doesn't succeed in the timeout
    period, then the test is made to fail with a "Thread still alive"
    message.
    """

    def __init__(self, testcase):
        threading.Thread.__init__(self)
        # In case this thread hangs, don't stop Python from exiting.
        self.daemon = True
        self._exc_info = None
        self._testcase = testcase

    def run(self):
        try:
            self.testrun()
        except:  # NOQA: E722 blank except
            self._exc_info = sys.exc_info()

    def cleanup(self, timeout=15):
        self.join(timeout)
        if self._exc_info:
            et, ev, etb = self._exc_info
            if ev is None:
                ev = et()
            raise ev.with_traceback(etb)
        if self.is_alive():
            self._testcase.fail("Thread did not finish: %s" % self)



##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Disk-based client cache for ZEO.

ClientCache exposes an API used by the ZEO client storage.  FileCache stores
objects on disk using a 2-tuple of oid and tid as key.

ClientCache's API is similar to a storage API, with methods like load(),
store(), and invalidate().  It manages in-memory data structures that allow
it to map this richer API onto the simple key-based API of the lower-level
FileCache.
"""
import logging
import os
import tempfile
import time
from struct import pack
from struct import unpack

import BTrees.LLBTree
import BTrees.LOBTree
import zc.lockfile
import ZODB.fsIndex
from ZODB.utils import RLock
from ZODB.utils import p64
from ZODB.utils import u64
from ZODB.utils import z64

from ._compat import PYPY


logger = logging.getLogger("ZEO.cache")

# A disk-based cache for ZEO clients.
#
# This class provides an interface to a persistent, disk-based cache
# used by ZEO clients to store copies of database records from the
# server.
#
# The details of the constructor as unspecified at this point.
#
# Each entry in the cache is valid for a particular range of transaction
# ids.  The lower bound is the transaction that wrote the data.  The
# upper bound is the next transaction that wrote a revision of the
# object.  If the data is current, the upper bound is stored as None;
# the data is considered current until an invalidate() call is made.
#
# It is an error to call store() twice with the same object without an
# intervening invalidate() to set the upper bound on the first cache
# entry.  Perhaps it will be necessary to have a call the removes
# something from the cache outright, without keeping a non-current
# entry.

# Cache verification
#
# When the client is connected to the server, it receives
# invalidations every time an object is modified.  When the client is
# disconnected then reconnects, it must perform cache verification to make
# sure its cached data is synchronized with the storage's current state.
#
# quick verification
# full verification
#


# FileCache stores a cache in a single on-disk file.
#
# On-disk cache structure.
#
# The file begins with a 12-byte header.  The first four bytes are the
# file's magic number - ZEC3 - indicating zeo cache version 4.  The
# next eight bytes are the last transaction id.

magic = b"ZEC3"
ZEC_HEADER_SIZE = 12

# Maximum block size. Note that while we are doing a store, we may
# need to write a free block that is almost twice as big.  If we die
# in the middle of a store, then we need to split the large free records
# while opening.
max_block_size = (1 << 31) - 1


# After the header, the file contains a contiguous sequence of blocks.  All
# blocks begin with a one-byte status indicator:
#
# b'a'
#       Allocated.  The block holds an object; the next 4 bytes are >I
#       format total block size.
#
# b'f'
#       Free.  The block is free; the next 4 bytes are >I format total
#       block size.
#
# b'1', b'2', b'3', b'4'
#       The block is free, and consists of 1, 2, 3 or 4 bytes total.
#
# "Total" includes the status byte, and size bytes.  There are no
# empty (size 0) blocks.


# Allocated blocks have more structure:
#
#     1 byte allocation status (b'a').
#     4 bytes block size, >I format.
#     8 byte oid
#     8 byte start_tid
#     8 byte end_tid
#     2 byte version length must be 0
#     4 byte data size
#     data
#     8 byte redundant oid for error detection.
allocated_record_overhead = 43

# The cache's currentofs goes around the file, circularly, forever.
# It's always the starting offset of some block.
#
# When a new object is added to the cache, it's stored beginning at
# currentofs, and currentofs moves just beyond it.  As many contiguous
# blocks needed to make enough room for the new object are evicted,
# starting at currentofs.  Exception:  if currentofs is close enough
# to the end of the file that the new object can't fit in one
# contiguous chunk, currentofs is reset to ZEC_HEADER_SIZE first.

# Under PyPy, the available dict specializations perform significantly
# better (faster) than the pure-Python BTree implementation. They may
# use less memory too. And we don't require any of the special BTree features.
_current_index_type = ZODB.fsIndex.fsIndex if not PYPY else dict
_noncurrent_index_type = BTrees.LOBTree.LOBTree if not PYPY else dict
# ...except at this leaf level
_noncurrent_bucket_type = BTrees.LLBTree.LLBucket


class ClientCache:
    """A simple in-memory cache."""

    # The default size of 200MB makes a lot more sense than the traditional
    # default of 20MB.  The default here is misleading, though, since
    # ClientStorage is the only user of ClientCache, and it always passes an
    # explicit size of its own choosing.
    def __init__(self, path=None, size=200*1024**2, rearrange=.8):

        # - `path`:  filepath for the cache file, or None (in which case
        #   a temp file will be created)
        self.path = path

        # - `maxsize`:  total size of the cache file
        #               We set to the minimum size of less than the minimum.
        size = max(size, ZEC_HEADER_SIZE)
        self.maxsize = size

        # rearrange: if we read a current record and it's more than
        # rearrange*size from the end, then copy it forward to keep it
        # from being evicted.
        self.rearrange = rearrange * size

        # The number of records in the cache.
        self._len = 0

        # {oid -> pos}
        self.current = _current_index_type()

        # {oid -> {tid->pos}}
        # Note that caches in the wild seem to have very little non-current
        # data, so this would seem to have little impact on memory consumption.
        # I wonder if we even need to store non-current data in the cache.
        self.noncurrent = _noncurrent_index_type()

        # tid for the most recent transaction we know about.  This is also
        # stored near the start of the file.
        self.tid = z64

        # Always the offset into the file of the start of a block.
        # New and relocated objects are always written starting at
        # currentofs.
        self.currentofs = ZEC_HEADER_SIZE

        self._lock = RLock()

        # self.f is the open file object.
        # When we're not reusing an existing file, self.f is left None
        # here -- the scan() method must be called then to open the file
        # (and it sets self.f).

        fsize = ZEC_HEADER_SIZE
        if path:
            self._lock_file = zc.lockfile.LockFile(path + '.lock')
            if not os.path.exists(path):
                # Create a small empty file. We'll make it bigger in _initfile.
                self.f = open(path, 'wb+')
                self.f.write(magic+z64)
                logger.info("created persistent cache file %r", path)
            else:
                fsize = os.path.getsize(self.path)
                self.f = open(path, 'rb+')
                logger.info("reusing persistent cache file %r", path)
        else:
            # Create a small empty file.  We'll make it bigger in _initfile.
            self.f = tempfile.TemporaryFile()
            self.f.write(magic+z64)
            logger.info("created temporary cache file %r", self.f.name)

        try:
            self._initfile(fsize)
        except:  # NOQA: E722 bare except
            self.f.close()
            if not path:
                raise  # unrecoverable temp file error :(
            badpath = path+'.bad'
            if os.path.exists(badpath):
                logger.critical(
                    'Removing bad cache file: %r (prev bad exists).',
                    path, exc_info=1)
                os.remove(path)
            else:
                logger.critical('Moving bad cache file to %r.',
                                badpath, exc_info=1)
                os.rename(path, badpath)
            self.f = open(path, 'wb+')
            self.f.write(magic+z64)
            self._initfile(ZEC_HEADER_SIZE)

        # Statistics:  _n_adds, _n_added_bytes,
        #              _n_evicts, _n_evicted_bytes,
        #              _n_accesses
        self.clearStats()

        self._setup_trace(path)

    # Backward compatibility. Client code used to have to use the fc
    # attr to get to the file cache to get cache stats.
    @property
    def fc(self):
        return self

    def clear(self):
        with self._lock:
            self.f.seek(ZEC_HEADER_SIZE)
            self.f.truncate()
            self._initfile(ZEC_HEADER_SIZE)

    ##
    # Scan the current contents of the cache file, calling `install`
    # for each object found in the cache.  This method should only
    # be called once to initialize the cache from disk.
    def _initfile(self, fsize):
        maxsize = self.maxsize
        f = self.f
        read = f.read
        seek = f.seek
        write = f.write
        seek(0)
        if read(4) != magic:
            seek(0)
            raise ValueError("unexpected magic number: %r" % read(4))
        self.tid = read(8)
        if len(self.tid) != 8:
            raise ValueError("cache file too small -- no tid at start")

        # Populate .filemap and .key2entry to reflect what's currently in the
        # file, and tell our parent about it too (via the `install` callback).
        # Remember the location of the largest free block.  That seems a
        # decent place to start currentofs.

        self.current = _current_index_type()
        self.noncurrent = _noncurrent_index_type()
        length = 0
        last = ofs = ZEC_HEADER_SIZE
        first_free_offset = 0
        current = self.current
        status = b' '
        while ofs < fsize:
            seek(ofs)
            status = read(1)
            if status == b'a':
                size, oid, start_tid, end_tid, lver = unpack(
                    ">I8s8s8sH", read(30))
                if ofs+size <= maxsize:
                    if end_tid == z64:
                        assert oid not in current, (ofs, f.tell())
                        current[oid] = ofs
                    else:
                        assert start_tid < end_tid, (ofs, f.tell())
                        self._set_noncurrent(oid, start_tid, ofs)
                    assert lver == 0, "Versions aren't supported"
                    length += 1
            else:
                # free block
                if first_free_offset == 0:
                    first_free_offset = ofs
                if status == b'f':
                    size, = unpack(">I", read(4))
                    if size > max_block_size:
                        # Oops, we either have an old cache, or a we
                        # crashed while storing. Split this block into two.
                        assert size <= max_block_size*2
                        seek(ofs+max_block_size)
                        write(b'f'+pack(">I", size-max_block_size))
                        seek(ofs)
                        write(b'f'+pack(">I", max_block_size))
                        sync(f)
                elif status in b'1234':
                    size = int(status)
                else:
                    raise ValueError("unknown status byte value %s in client "
                                     "cache file" % 0, hex(ord(status)))

            last = ofs
            ofs += size

            if ofs >= maxsize:
                # Oops, the file was bigger before.
                if ofs > maxsize:
                    # The last record is too big. Replace it with a smaller
                    # free record
                    size = maxsize-last
                    seek(last)
                    if size > 4:
                        write(b'f'+pack(">I", size))
                    else:
                        write("012345"[size].encode())
                    sync(f)
                    ofs = maxsize
                break

        if fsize < maxsize:
            assert ofs == fsize
            # Make sure the OS really saves enough bytes for the file.
            seek(self.maxsize - 1)
            write(b'x')

            # add as many free blocks as are needed to fill the space
            seek(ofs)
            nfree = maxsize - ZEC_HEADER_SIZE
            for i in range(0, nfree, max_block_size):
                block_size = min(max_block_size, nfree-i)
                write(b'f' + pack(">I", block_size))
                seek(block_size-5, 1)
            sync(self.f)

            # There is always data to read and
            assert last and (status in b' f1234')
            first_free_offset = last
        else:
            assert ofs == maxsize
            if maxsize < fsize:
                seek(maxsize)
                f.truncate()

        # We use the first_free_offset because it is most likely the
        # place where we last wrote.
        self.currentofs = first_free_offset or ZEC_HEADER_SIZE
        self._len = length

    def _set_noncurrent(self, oid, tid, ofs):
        noncurrent_for_oid = self.noncurrent.get(u64(oid))
        if noncurrent_for_oid is None:
            noncurrent_for_oid = _noncurrent_bucket_type()
            self.noncurrent[u64(oid)] = noncurrent_for_oid
        noncurrent_for_oid[u64(tid)] = ofs

    def _del_noncurrent(self, oid, tid):
        try:
            noncurrent_for_oid = self.noncurrent[u64(oid)]
            del noncurrent_for_oid[u64(tid)]
            if not noncurrent_for_oid:
                del self.noncurrent[u64(oid)]
        except KeyError:
            logger.error("Couldn't find non-current %r", (oid, tid))

    def clearStats(self):
        self._n_adds = self._n_added_bytes = 0
        self._n_evicts = self._n_evicted_bytes = 0
        self._n_accesses = 0

    def getStats(self):
        return (self._n_adds, self._n_added_bytes,
                self._n_evicts, self._n_evicted_bytes,
                self._n_accesses)

    ##
    # The number of objects currently in the cache.
    def __len__(self):
        return self._len

    ##
    # Close the underlying file.  No methods accessing the cache should be
    # used after this.
    def close(self):
        self._unsetup_trace()
        f = self.f
        self.f = None
        if f is not None:
            sync(f)
            f.close()

        if hasattr(self, '_lock_file'):
            self._lock_file.close()

    ##
    # Evict objects as necessary to free up at least nbytes bytes,
    # starting at currentofs.  If currentofs is closer than nbytes to
    # the end of the file, currentofs is reset to ZEC_HEADER_SIZE first.
    # The number of bytes actually freed may be (and probably will be)
    # greater than nbytes, and is _makeroom's return value.  The file is not
    # altered by _makeroom.  filemap and key2entry are updated to reflect the
    # evictions, and it's the caller's responsibility both to fiddle
    # the file, and to update filemap, to account for all the space
    # freed (starting at currentofs when _makeroom returns, and
    # spanning the number of bytes retured by _makeroom).
    def _makeroom(self, nbytes):
        assert 0 < nbytes <= self.maxsize - ZEC_HEADER_SIZE, (
            nbytes, self.maxsize)
        if self.currentofs + nbytes > self.maxsize:
            self.currentofs = ZEC_HEADER_SIZE
        ofs = self.currentofs
        seek = self.f.seek
        read = self.f.read
        current = self.current
        while nbytes > 0:
            seek(ofs)
            status = read(1)
            if status == b'a':
                size, oid, start_tid, end_tid = unpack(">I8s8s8s", read(28))
                self._n_evicts += 1
                self._n_evicted_bytes += size
                if end_tid == z64:
                    del current[oid]
                else:
                    self._del_noncurrent(oid, start_tid)
                self._len -= 1
            else:
                if status == b'f':
                    size = unpack(">I", read(4))[0]
                else:
                    assert status in b'1234'
                    size = int(status)
            ofs += size
            nbytes -= size
        return ofs - self.currentofs

    ##
    # Update our idea of the most recent tid.  This is stored in the
    # instance, and also written out near the start of the cache file.  The
    # new tid must be strictly greater than our current idea of the most
    # recent tid.
    def setLastTid(self, tid):
        with self._lock:
            if (not tid) or (tid == z64):
                return
            if (tid <= self.tid) and self._len:
                if tid == self.tid:
                    return                  # Be a little forgiving
                raise ValueError("new last tid (%s) must be greater than "
                                 "previous one (%s)"
                                 % (u64(tid), u64(self.tid)))
            assert isinstance(tid, bytes) and len(tid) == 8, tid
            self.tid = tid
            self.f.seek(len(magic))
            self.f.write(tid)
            self.f.flush()

    ##
    # Return the last transaction seen by the cache.
    # @return a transaction id
    # @defreturn string, or 8 nulls if no transaction is yet known
    def getLastTid(self):
        with self._lock:
            return self.tid

    ##
    # Return the current data record for oid.
    # @param oid object id
    # @return (data record, serial number, tid), or None if the object is not
    #         in the cache
    # @defreturn 3-tuple: (string, string, string)
    def load(self, oid, before_tid=None):
        with self._lock:
            ofs = self.current.get(oid)
            if ofs is None:
                self._trace(0x20, oid)
                return None
            self.f.seek(ofs)
            read = self.f.read
            status = read(1)
            assert status == b'a', (ofs, self.f.tell(), oid)
            size, saved_oid, tid, end_tid, lver, ldata = unpack(
                ">I8s8s8sHI", read(34))
            assert saved_oid == oid, (ofs, self.f.tell(), oid, saved_oid)
            assert end_tid == z64, (ofs, self.f.tell(), oid, tid, end_tid)
            assert lver == 0, "Versions aren't supported"

            if before_tid and tid >= before_tid:
                return None

            data = read(ldata)
            assert len(data) == ldata, (
                ofs, self.f.tell(), oid, len(data), ldata)

            # WARNING: The following assert changes the file position.
            # We must not depend on this below or we'll fail in optimized mode.
            assert read(8) == oid, (ofs, self.f.tell(), oid)

            self._n_accesses += 1
            self._trace(0x22, oid, tid, end_tid, ldata)

            ofsofs = self.currentofs - ofs
            if ofsofs < 0:
                ofsofs += self.maxsize

            if ofsofs > self.rearrange and \
               self.maxsize > 10*len(data) and \
               size > 4:
                # The record is far back and might get evicted, but it's
                # valuable, so move it forward.

                # Remove fromn old loc:
                del self.current[oid]
                self.f.seek(ofs)
                self.f.write(b'f'+pack(">I", size))

                # Write to new location:
                self._store(oid, tid, None, data, size)

            return data, tid

    ##
    # Return a non-current revision of oid that was current before tid.
    # @param oid object id
    # @param tid id of transaction that wrote next revision of oid
    # @return data record, serial number, start tid, and end tid
    # @defreturn 4-tuple: (string, string, string, string)
    def loadBefore(self, oid, before_tid):
        with self._lock:
            noncurrent_for_oid = self.noncurrent.get(u64(oid))
            if noncurrent_for_oid is None:
                result = self.load(oid, before_tid)
                if result:
                    return result[0], result[1], None
                else:
                    self._trace(0x24, oid, "", before_tid)
                    return result

            items = noncurrent_for_oid.items(None, u64(before_tid)-1)
            if not items:
                result = self.load(oid, before_tid)
                if result:
                    return result[0], result[1], None
                else:
                    self._trace(0x24, oid, "", before_tid)
                    return result

            tid, ofs = items[-1]

            self.f.seek(ofs)
            read = self.f.read
            status = read(1)
            assert status == b'a', (ofs, self.f.tell(), oid, before_tid)
            size, saved_oid, saved_tid, end_tid, lver, ldata = unpack(
                ">I8s8s8sHI", read(34))
            assert saved_oid == oid, (ofs, self.f.tell(), oid, saved_oid)
            assert saved_tid == p64(tid), (
                ofs, self.f.tell(), oid, saved_tid, tid)
            assert end_tid != z64, (ofs, self.f.tell(), oid)
            assert lver == 0, "Versions aren't supported"
            data = read(ldata)
            assert len(data) == ldata, (ofs, self.f.tell())

            # WARNING: The following assert changes the file position.
            # We must not depend on this below or we'll fail in optimized mode.
            assert read(8) == oid, (ofs, self.f.tell(), oid)

            if end_tid < before_tid:
                result = self.load(oid, before_tid)
                if result:
                    return result[0], result[1], None
                else:
                    self._trace(0x24, oid, "", before_tid)
                    return result

            self._n_accesses += 1
            self._trace(0x26, oid, "", saved_tid)
            return data, saved_tid, end_tid

    ##
    # Store a new data record in the cache.
    # @param oid object id
    # @param start_tid the id of the transaction that wrote this revision
    # @param end_tid the id of the transaction that created the next
    #                revision of oid.  If end_tid is None, the data is
    #                current.
    # @param data the actual data
    def store(self, oid, start_tid, end_tid, data):
        with self._lock:
            seek = self.f.seek
            if end_tid is None:
                ofs = self.current.get(oid)
                if ofs:
                    seek(ofs)
                    read = self.f.read
                    status = read(1)
                    assert status == b'a', (ofs, self.f.tell(), oid)
                    size, saved_oid, saved_tid, end_tid = unpack(
                        ">I8s8s8s", read(28))
                    assert saved_oid == oid, (
                        ofs, self.f.tell(), oid, saved_oid)
                    assert end_tid == z64, (ofs, self.f.tell(), oid)
                    if saved_tid == start_tid:
                        return
                    raise ValueError("already have current data for oid")
            else:
                noncurrent_for_oid = self.noncurrent.get(u64(oid))
                if noncurrent_for_oid and \
                   u64(start_tid) in noncurrent_for_oid:
                    return

            size = allocated_record_overhead + len(data)

            # A number of cache simulation experiments all concluded that the
            # 2nd-level ZEO cache got a much higher hit rate if "very large"
            # objects simply weren't cached.  For now, we ignore the request
            # only if the entire cache file is too small to hold the object.
            if size >= min(max_block_size, self.maxsize - ZEC_HEADER_SIZE):
                return

            self._n_adds += 1
            self._n_added_bytes += size
            self._len += 1

            self._store(oid, start_tid, end_tid, data, size)

            if end_tid:
                self._trace(0x54, oid, start_tid, end_tid, dlen=len(data))
            else:
                self._trace(0x52, oid, start_tid, dlen=len(data))

    def _store(self, oid, start_tid, end_tid, data, size):
        # Low-level store used by store and load

        # In the next line, we ask for an extra to make sure we always
        # have a free block after the new alocated block.  This free
        # block acts as a ring pointer, so that on restart, we start
        # where we left off.
        nfreebytes = self._makeroom(size+1)

        assert size <= nfreebytes, (size, nfreebytes)
        excess = nfreebytes - size
        # If there's any excess (which is likely), we need to record a
        # free block following the end of the data record.  That isn't
        # expensive -- it's all a contiguous write.
        if excess == 0:
            extra = b''
        elif excess < 5:
            extra = "01234"[excess].encode()
        else:
            extra = b'f' + pack(">I", excess)

        ofs = self.currentofs
        seek = self.f.seek
        seek(ofs)
        write = self.f.write

        # Before writing data, we'll write a free block for the space freed.
        # We'll come back with a last atomic write to rewrite the start of the
        # allocated-block header.
        write(b'f'+pack(">I", nfreebytes))

        # Now write the rest of the allocation block header and object data.
        write(pack(">8s8s8sHI", oid, start_tid, end_tid or z64, 0, len(data)))
        write(data)
        write(oid)
        write(extra)

        # Now, we'll go back and rewrite the beginning of the
        # allocated block header.
        seek(ofs)
        write(b'a'+pack(">I", size))

        if end_tid:
            self._set_noncurrent(oid, start_tid, ofs)
        else:
            self.current[oid] = ofs

        self.currentofs += size

    ##
    # If `tid` is None,
    # forget all knowledge of `oid`.  (`tid` can be None only for
    # invalidations generated by startup cache verification.)  If `tid`
    # isn't None, and we had current
    # data for `oid`, stop believing we have current data, and mark the
    # data we had as being valid only up to `tid`.  In all other cases, do
    # nothing.
    #
    # Paramters:
    #
    # - oid object id
    # - tid the id of the transaction that wrote a new revision of oid,
    #        or None to forget all cached info about oid.
    def invalidate(self, oid, tid):
        with self._lock:
            ofs = self.current.get(oid)
            if ofs is None:
                # 0x10 == invalidate (miss)
                self._trace(0x10, oid, tid)
                return

            self.f.seek(ofs)
            read = self.f.read
            status = read(1)
            assert status == b'a', (ofs, self.f.tell(), oid)
            size, saved_oid, saved_tid, end_tid = unpack(">I8s8s8s", read(28))
            assert saved_oid == oid, (ofs, self.f.tell(), oid, saved_oid)
            assert end_tid == z64, (ofs, self.f.tell(), oid)
            del self.current[oid]
            if tid is None:
                self.f.seek(ofs)
                self.f.write(b'f'+pack(">I", size))
                # 0x1E = invalidate (hit, discarding current or non-current)
                self._trace(0x1E, oid, tid)
                self._len -= 1
            else:
                if tid == saved_tid:
                    logger.warning(
                        "Ignoring invalidation with same tid as current")
                    return
                self.f.seek(ofs+21)
                self.f.write(tid)
                self._set_noncurrent(oid, saved_tid, ofs)
                # 0x1C = invalidate (hit, saving non-current)
                self._trace(0x1C, oid, tid)

    ##
    # Generates (oid, serial) oairs for all objects in the
    # cache.  This generator is used by cache verification.
    def contents(self):
        # May need to materialize list instead of iterating;
        # depends on whether the caller may change the cache.
        seek = self.f.seek
        read = self.f.read
        for oid, ofs in self.current.items():
            seek(ofs)
            status = read(1)
            assert status == b'a', (ofs, self.f.tell(), oid)
            size, saved_oid, tid, end_tid = unpack(">I8s8s8s", read(28))
            assert saved_oid == oid, (ofs, self.f.tell(), oid, saved_oid)
            assert end_tid == z64, (ofs, self.f.tell(), oid)
            yield oid, tid

    def dump(self):
        from ZODB.utils import oid_repr
        print("cache size", len(self))
        L = list(self.contents())
        L.sort()
        for oid, tid in L:
            print(oid_repr(oid), oid_repr(tid))
        print("dll contents")
        L = sorted(list(self), key=lambda x: x.key)
        for x in L:
            end_tid = x.end_tid or z64
            print(oid_repr(x.key[0]), oid_repr(x.key[1]), oid_repr(end_tid))
        print()

    # If `path` isn't None (== we're using a persistent cache file), and
    # envar ZEO_CACHE_TRACE is set to a non-empty value, try to open
    # path+'.trace' as a trace file, and store the file object in
    # self._tracefile.  If not, or we can't write to the trace file, disable
    # tracing by setting self._trace to a dummy function, and set
    # self._tracefile to None.
    _tracefile = None

    def _trace(self, *a, **kw):
        pass

    def _setup_trace(self, path):
        _tracefile = None
        if path and os.environ.get("ZEO_CACHE_TRACE"):
            tfn = path + ".trace"
            try:
                _tracefile = open(tfn, "ab")
            except OSError as msg:
                logger.warning("cannot write tracefile %r (%s)", tfn, msg)
            else:
                logger.info("opened tracefile %r", tfn)

        if _tracefile is None:
            return

        now = time.time

        def _trace(code, oid=b"", tid=z64, end_tid=z64, dlen=0):
            # The code argument is two hex digits; bits 0 and 7 must be zero.
            # The first hex digit shows the operation, the second the outcome.
            # This method has been carefully tuned to be as fast as possible.
            # Note: when tracing is disabled, this method is hidden by a dummy.
            encoded = (dlen << 8) + code
            if tid is None:
                tid = z64
            if end_tid is None:
                end_tid = z64
            try:
                _tracefile.write(
                    pack(">iiH8s8s",
                         int(now()), encoded, len(oid), tid, end_tid) + oid,
                    )
            except:  # NOQA: E722 bare except
                print(repr(tid), repr(end_tid))
                raise

        self._trace = _trace
        self._tracefile = _tracefile
        _trace(0x00)

    def _unsetup_trace(self):
        if self._tracefile is not None:
            del self._trace
            self._tracefile.close()
            del self._tracefile


def sync(f):
    f.flush()
    os.fsync(f.fileno())



#!/usr/bin/env python2.3

"""Transaction timeout test script.

This script connects to a storage, begins a transaction, calls store()
and tpc_vote(), and then sleeps forever.  This should trigger the
transaction timeout feature of the server.

usage: timeout.py address delay [storage-name]

"""

import sys
import time

from ZODB.Connection import TransactionMetaData
from ZODB.tests.MinPO import MinPO
from ZODB.tests.StorageTestBase import zodb_pickle

from ZEO.ClientStorage import ClientStorage


ZERO = '\0'*8


def main():
    if len(sys.argv) not in (3, 4):
        sys.stderr.write("Usage: timeout.py address delay [storage-name]\n" %
                         sys.argv[0])
        sys.exit(2)

    hostport = sys.argv[1]
    delay = float(sys.argv[2])
    if sys.argv[3:]:
        name = sys.argv[3]
    else:
        name = "1"

    if "/" in hostport:
        address = hostport
    else:
        if ":" in hostport:
            i = hostport.index(":")
            host, port = hostport[:i], hostport[i+1:]
        else:
            host, port = "", hostport
        port = int(port)
        address = (host, port)

    print("Connecting to %s..." % repr(address))
    storage = ClientStorage(address, name)
    print("Connected.  Now starting a transaction...")

    oid = storage.new_oid()
    revid = ZERO
    data = MinPO("timeout.py")
    pickled_data = zodb_pickle(data)
    t = TransactionMetaData()
    t.user = "timeout.py"
    storage.tpc_begin(t)
    storage.store(oid, revid, pickled_data, '', t)
    print("Stored.  Now voting...")
    storage.tpc_vote(t)

    print("Voted; now sleeping %s..." % delay)
    time.sleep(delay)
    print("Done.")


if __name__ == "__main__":
    main()



##############################################################################
#
# Copyright (c) 2006 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import doctest
import unittest


class FakeStorageBase:

    def __getattr__(self, name):
        if name in ('getTid', 'history', 'load', 'loadSerial',
                    'lastTransaction', 'getSize', 'getName', 'supportsUndo',
                    'tpc_transaction'):
            return lambda *a, **k: None
        raise AttributeError(name)

    def isReadOnly(self):
        return False

    def __len__(self):
        return 4


class FakeStorage(FakeStorageBase):

    def record_iternext(self, next=None):
        if next is None:
            next = '0'
        next = str(int(next) + 1)
        oid = next
        if next == '4':
            next = None

        return oid, oid*8, 'data ' + oid, next


class FakeServer:
    storages = {
        '1': FakeStorage(),
        '2': FakeStorageBase(),
        }
    lock_managers = storages

    def register_connection(*args):
        return None, None

    client_conflict_resolution = False


class FakeConnection:
    protocol_version = b'Z5'
    addr = 'test'

    def call_soon_threadsafe(f, *a):
        return f(*a)

    async_ = async_threadsafe = None


def test_server_record_iternext():
    """

On the server, record_iternext calls are simply delegated to the
underlying storage.

    >>> import ZEO.StorageServer

    >>> zeo = ZEO.StorageServer.ZEOStorage(FakeServer(), False)
    >>> zeo.notify_connected(FakeConnection())
    >>> zeo.register('1', False)

    >>> next = None
    >>> while 1:
    ...     oid, serial, data, next = zeo.record_iternext(next)
    ...     print(oid)
    ...     if next is None:
    ...         break
    1
    2
    3
    4

The storage info also reflects the fact that record_iternext is supported.

    >>> zeo.get_info()['supports_record_iternext']
    True

    >>> zeo = ZEO.StorageServer.ZEOStorage(FakeServer(), False)
    >>> zeo.notify_connected(FakeConnection())
    >>> zeo.register('2', False)

    >>> zeo.get_info()['supports_record_iternext']
    False

"""


def test_client_record_iternext():
    """Test client storage delegation to the network client

The client simply delegates record_iternext calls to it's server stub.

There's really no decent way to test ZEO without running too much crazy
stuff.  I'd rather do a lame test than a really lame test, so here goes.

First, fake out the connection manager so we can make a connection:

    >>> import ZEO

    >>> class Client(ZEO.asyncio.testing.ClientRunner):
    ...
    ...    def record_iternext(self, next=None):
    ...        if next == None:
    ...           next = '0'
    ...        next = str(int(next) + 1)
    ...        oid = next
    ...        if next == '4':
    ...            next = None
    ...
    ...        return oid, oid*8, 'data ' + oid, next
    ...

    >>> client = ZEO.client(
    ...     '', wait=False, _client_factory=Client)

Now we'll have our way with it's private _server attr:

    >>> next = None
    >>> while 1:
    ...     oid, serial, data, next = client.record_iternext(next)
    ...     print(oid)
    ...     if next is None:
    ...         break
    1
    2
    3
    4
    >>> client.close()

"""


def test_suite():
    return doctest.DocTestSuite()



##############################################################################
#
# Copyright (c) 2022 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Module racetest provides infrastructure for tests that catch concurrency
problems in ZEO.

It complements ZODB.tests.racetest .
"""

import time

import zope.interface


# LoadDelayedStorage wraps base storage in injects delays after load*
# operations.
#
# It is useful to catch concurrency problems due to races in between load and
# other storage events, e.g. invalidations.
#
# See e.g. https://github.com/zopefoundation/ZEO/issues/209 for example where
# injecting such delays was effective to catch bugs that were otherwise hard to
# reproduce.
class LoadDelayedStorage:

    def __init__(self, base, tdelay=0.01):
        self.base = base
        self.tdelay = tdelay
        zope.interface.directlyProvides(self, zope.interface.providedBy(base))

    def _delay(self):
        time.sleep(self.tdelay)

    def __getattr__(self, name):
        return getattr(self.base, name)

    def __len__(self):          # ^^^ __getattr__ does not forward __<attr>
        return len(self.base)   # because they are mangled

    def load(self, *argv, **kw):
        _ = self.base.load(*argv, **kw)
        self._delay()
        return _

    def loadBefore(self, *argv, **kw):
        _ = self.base.loadBefore(*argv, **kw)
        self._delay()
        return _

    def loadSerial(self, *argv, **kw):
        _ = self.base.loadSerial(*argv, **kw)
        self._delay()
        return _


class ZConfigLoadDelayed:

    _factory = LoadDelayedStorage

    def __init__(self, config):
        self.config = config
        self.name = config.getSectionName()

    def open(self):
        base = self.config.base.open()
        return self._factory(base)



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Library for forking storage server and connecting client storage"""
import gc
import logging
import multiprocessing
import os
import os.path
import sys
import tempfile
from io import StringIO
from queue import Empty


logger = logging.getLogger('ZEO.tests.forker')
DEBUG = os.environ.get('ZEO_TEST_SERVER_DEBUG')


class ZEOConfig:
    """Class to generate ZEO configuration file. """

    def __init__(self, addr, log=None, **options):
        if log:
            if isinstance(log, str):
                self.logpath = log
            elif isinstance(addr, str):
                self.logpath = addr+'.log'
            else:
                self.logpath = 'server.log'

        if not isinstance(addr, str):
            addr = '%s:%s' % addr

        self.log = log
        self.address = addr
        self.read_only = None
        self.loglevel = 'INFO'
        self.__dict__.update(options)

    def dump(self, f):
        print("<zeo>", file=f)
        print("address " + self.address, file=f)
        if self.read_only is not None:
            print("read-only", self.read_only and "true" or "false", file=f)

        for name in (
                'invalidation_queue_size', 'invalidation_age',
                'transaction_timeout', 'pid_filename', 'msgpack',
                'ssl_certificate', 'ssl_key', 'client_conflict_resolution'):
            v = getattr(self, name, None)
            if v:
                print(name.replace('_', '-'), v, file=f)

        print("</zeo>", file=f)

        if self.log:
            print(f"""
            <eventlog>
              level {self.loglevel}
              <logfile>
                 path {self.logpath}
              </logfile>
            </eventlog>
            """, file=f)

    def __str__(self):
        f = StringIO()
        self.dump(f)
        return f.getvalue()


def runner(config, qin, qout, timeout=None,
           debug=False, name=None,
           keep=False, protocol=None):

    root_logger = logging.getLogger()
    if not root_logger.handlers:
        # logging is not configured -- maybe because
        # the process has been ``spawn``ed.
        # Try to recreate the logging configuration
        logini = os.environ.get("ZOPE_TESTRUNNER_LOG_INI")
        if logini is not None and os.path.exists(logini):
            from logging.config import fileConfig
            fileConfig(logini, disable_existing_loggers=False)

    if debug or DEBUG:
        debug_logging()

    old_protocol = None
    if protocol:
        import ZEO.asyncio.server
        old_protocol = ZEO.asyncio.server.best_protocol_version
        ZEO.asyncio.server.best_protocol_version = protocol
        old_protocols = ZEO.asyncio.server.ServerProtocol.protocols
        ZEO.asyncio.server.ServerProtocol.protocols = tuple(sorted(
            set(old_protocols) | {protocol}
            ))

    try:
        import threading

        from . import runzeo

        options = runzeo.ZEOOptions()
        options.realize(['-C', config])
        server = runzeo.ZEOServer(options)
        globals()[(name if name else 'last') + '_server'] = server
        server.open_storages()
        server.clear_socket()
        server.create_server()
        logger.debug('SERVER CREATED')
        qout.put(server.server.acceptor.addr)
        logger.debug('ADDRESS SENT')
        thread = threading.Thread(
            target=server.server.loop, kwargs=dict(timeout=.2),
            name=(None if name is None else name + '-server'),
            )
        thread.daemon = True
        thread.start()
        os.remove(config)

        try:
            qin.get(timeout=timeout)  # wait for shutdown
        except Empty:
            pass
        server.server.close()
        thread.join(3)

        if not keep:
            # Try to cleanup storage files
            for storage in server.server.storages.values():
                try:
                    storage.cleanup()
                except AttributeError:
                    pass

        qout.put(thread.is_alive())

    except Exception:
        logger.exception("In server thread")

    finally:
        if old_protocol:
            ZEO.asyncio.server.best_protocol_version = old_protocol
            ZEO.asyncio.server.ServerProtocol.protocols = old_protocols


def stop_runner(thread, config, qin, qout, stop_timeout=19, pid=None):
    qin.put('stop')
    try:
        dirty = qout.get(timeout=stop_timeout)
    except Empty:
        print("WARNING Couldn't stop server", file=sys.stderr)
        if hasattr(thread, 'terminate'):
            thread.terminate()
            os.waitpid(thread.pid, 0)
    else:
        if dirty:
            print("WARNING SERVER DIDN'T STOP CLEANLY", file=sys.stderr)

            # The runner thread didn't stop. If it was a process,
            # give it some time to exit
            if hasattr(thread, 'pid') and thread.pid:
                os.waitpid(thread.pid, 0)

    thread.join(stop_timeout)

    gc.collect()


def start_zeo_server(storage_conf=None, zeo_conf=None, port=None, keep=False,
                     path='Data.fs', protocol=None, blob_dir=None,
                     suicide=True, debug=False,
                     threaded=False, start_timeout=33, name=None, log=None,
                     show_config=False):
    """Start a ZEO server in a separate process.

    Takes two positional arguments a string containing the storage conf
    and a ZEOConfig object.

    Returns the ZEO address, the test server address, the pid, and the path
    to the config file.
    """

    if not storage_conf:
        storage_conf = '<filestorage>\npath %s\n</filestorage>' % path

    if blob_dir:
        storage_conf = (f'<blobstorage>\nblob-dir {blob_dir}\n'
                        f'{storage_conf}\n</blobstorage>')

    if zeo_conf is None or isinstance(zeo_conf, dict):
        if port is None:
            port = 0

        if isinstance(port, int):
            addr = '127.0.0.1', port
        else:
            addr = port

        z = ZEOConfig(addr, log=log)
        if zeo_conf:
            z.__dict__.update(zeo_conf)
        zeo_conf = str(z)

    zeo_conf = str(zeo_conf) + '\n\n' + storage_conf
    if show_config:
        print(zeo_conf)

    # Store the config info in a temp file.
    fd, tmpfile = tempfile.mkstemp(".conf", prefix='ZEO_forker',
                                   dir=os.getcwd())
    with os.fdopen(fd, 'w') as fp:
        fp.write(zeo_conf)

    if threaded:
        from queue import Queue
        from threading import Thread
    else:
        # Experimental to see whether server logging problems under MacOS
        # have to do with its default `spawn` method
        # from multiprocessing import Process as Thread
        process_type = os.environ.get("ZEO_PROCESS_TYPE", "")
        from multiprocessing import context
        Thread = getattr(context, process_type.capitalize() + "Process")
        Queue = ThreadlessQueue

    logger.info("using thread type %r", Thread)
    qin = Queue()
    qout = Queue()
    thread = Thread(
        target=runner,
        args=[tmpfile, qin, qout, 999 if suicide else None],
        kwargs=dict(debug=debug, name=name, protocol=protocol, keep=keep),
        name=(None if name is None else name + '-server-runner'),
    )
    thread.daemon = True
    thread.start()
    try:
        addr = qout.get(timeout=start_timeout)
    except Exception:
        whine("SERVER FAILED TO START")
        if thread.is_alive():
            whine("Server thread/process is still running")
        elif not threaded:
            whine("Exit status", thread.exitcode)
        raise

    def stop(stop_timeout=99):
        stop_runner(thread, tmpfile, qin, qout, stop_timeout)

    return addr, stop


def shutdown_zeo_server(stop):
    stop()


def debug_logging(logger='ZEO', stream='stderr', level=logging.DEBUG):
    handler = logging.StreamHandler(getattr(sys, stream))
    logger = logging.getLogger(logger)
    logger.addHandler(handler)
    logger.setLevel(level)

    def stop():
        logger.removeHandler(handler)
        logger.setLevel(logging.NOTSET)

    return stop


def whine(*message):
    print(*message, file=sys.stderr)
    sys.stderr.flush()


class ThreadlessQueue:

    def __init__(self):
        self.cin, self.cout = multiprocessing.Pipe(False)

    def put(self, v):
        self.cout.send(v)

    def get(self, timeout=None):
        if self.cin.poll(timeout):
            return self.cin.recv()
        else:
            raise Empty()



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

"""Test suite for ZEO.runzeo.ZEOOptions."""

import os
import tempfile
import unittest

import ZODB.config
from zdaemon.tests.testzdoptions import TestZDOptions

from ZEO.runzeo import ZEOOptions


# When a hostname isn't specified in a socket binding address, ZConfig
# supplies the empty string.
DEFAULT_BINDING_HOST = ""


class TestZEOOptions(TestZDOptions):

    OptionsClass = ZEOOptions

    input_args = ["-f", "Data.fs", "-a", "5555"]
    output_opts = [("-f", "Data.fs"), ("-a", "5555")]
    output_args = []

    configdata = """
        <zeo>
          address 5555
        </zeo>
        <filestorage fs>
          path Data.fs
        </filestorage>
        """

    def setUp(self):
        self.tempfilename = tempfile.mktemp()
        with open(self.tempfilename, "w") as f:
            f.write(self.configdata)

    def tearDown(self):
        try:
            os.remove(self.tempfilename)
        except OSError:
            pass

    def test_configure(self):
        # Hide the base class test_configure
        pass

    def test_default_help(self): pass  # disable silly test w spurious failures

    def test_defaults_with_schema(self):
        options = self.OptionsClass()
        options.realize(["-C", self.tempfilename])
        self.assertEqual(options.address, (DEFAULT_BINDING_HOST, 5555))
        self.assertEqual(len(options.storages), 1)
        opener = options.storages[0]
        self.assertEqual(opener.name, "fs")
        self.assertEqual(opener.__class__, ZODB.config.FileStorage)
        self.assertEqual(options.read_only, 0)
        self.assertEqual(options.transaction_timeout, None)
        self.assertEqual(options.invalidation_queue_size, 100)

    def test_defaults_without_schema(self):
        options = self.OptionsClass()
        options.realize(["-a", "5555", "-f", "Data.fs"])
        self.assertEqual(options.address, (DEFAULT_BINDING_HOST, 5555))
        self.assertEqual(len(options.storages), 1)
        opener = options.storages[0]
        self.assertEqual(opener.name, "1")
        self.assertEqual(opener.__class__, ZODB.config.FileStorage)
        self.assertEqual(opener.config.path, "Data.fs")
        self.assertEqual(options.read_only, 0)
        self.assertEqual(options.transaction_timeout, None)
        self.assertEqual(options.invalidation_queue_size, 100)

    def test_commandline_overrides(self):
        options = self.OptionsClass()
        options.realize(["-C", self.tempfilename,
                         "-a", "6666", "-f", "Wisdom.fs"])
        self.assertEqual(options.address, (DEFAULT_BINDING_HOST, 6666))
        self.assertEqual(len(options.storages), 1)
        opener = options.storages[0]
        self.assertEqual(opener.__class__, ZODB.config.FileStorage)
        self.assertEqual(opener.config.path, "Wisdom.fs")
        self.assertEqual(options.read_only, 0)
        self.assertEqual(options.transaction_timeout, None)
        self.assertEqual(options.invalidation_queue_size, 100)


del TestZDOptions  # don't run ZDaemon tests
